"""This module contains STIS/CCD commissioning tests.
Adapted from ``astrolib/pysynphot/from_commissioning/stis/test*.py``.
"""

# STDLIB
import os
import shutil

# THIRD-PARTY
from astropy.utils.data import get_pkg_data_filename

# LOCAL
from ..utils import CommCase

# Local test data
datafiles = ['earthshine.fits', 'el1215a.fits', 'el1302a.fits', 'el1356a.fits',
             'el2471a.fits', 'qso_template.fits', 'Zodi.fits']


def setup_module(module):
    """Copy test data to working directory so ``parse_spec`` works
    properly for both ``stsynphot`` and ASTROLIB PYSYNPHOT."""
    for datafile in datafiles:
        src = get_pkg_data_filename(os.path.join('data', datafile))
        shutil.copyfile(src, datafile)


def teardown_module(module):
    """Clean up test data in working directory."""
    for datafile in datafiles:
        os.remove(datafile)


class Test1101(CommCase):
    obsmode = 'stis,ccd'
    spectrum = 'rn(unit(1,flam),band(johnson,v),15.0,vegamag)'


class Test1102(CommCase):
    obsmode = 'stis,ccd,50ccd'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),10,vegamag)'


class Test1103(Test1102):
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),20,vegamag)'


class Test1104(Test1102):
    spectrum = 'rn(icat(k93models,5770,0.0,4.5),band(johnson,v),28,vegamag)'


class Test1105(Test1102):
    spectrum = ('rn(spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits),'
                'band(johnson,v),10,vegamag)')
    force = 'extrap'


class Test1106(Test1102):
    spectrum = 'rn(unit(1.0,flam),band(johnson,v),15,vegamag)'


class Test1107(Test1102):
    spectrum = 'spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits)'
    force = 'extrap'


class Test1108(Test1102):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test1109(Test1102):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1110(CommCase):
    obsmode = 'stis,ccd,f25nd5'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),4,vegamag)'


class Test1111(Test1110):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1112(CommCase):
    obsmode = 'stis,ccd,f28x50lp'
    spectrum = 'rn(icat(k93models,5860,0.0,4.4),band(johnson,v),5,vegamag)'


class Test1113(Test1112):
    spectrum = 'spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits)'
    force = 'extrap'


class Test1114(CommCase):
    obsmode = 'stis,ccd,f28x50lp'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test1115(Test1114):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1116(CommCase):
    obsmode = 'stis,ccd,f28x50oii'
    spectrum = 'spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits)'


class Test1117(Test1116):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test1118(CommCase):
    obsmode = 'stis,ccd,f28x50oiii'
    spectrum = 'spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits)'
    force = 'extrap'


class Test1119(CommCase):
    obsmode = 'stis,ccd,f28x50oiii'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test1120(CommCase):
    obsmode = 'stis,ccd,g230lb'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test1121(Test1120):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1122(CommCase):
    obsmode = 'stis,ccd,g230lb,s52x2'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),22,vegamag)'


class Test1123(Test1122):
    spectrum = 'spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits)'


class Test1124(CommCase):
    obsmode = 'stis,ccd,g230mb,c1995'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1125(CommCase):
    obsmode = 'stis,ccd,g230mb,c1995,s52x2'
    spectrum = 'spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits)'


class Test1126(CommCase):
    obsmode = 'stis,ccd,g430l'
    spectrum = 'em(4300.0,1.0,1.0E-12,flam)'


class Test1127(Test1126):
    spectrum = 'rn(icat(k93models,5860,0.0,4.4),band(johnson,v),5,vegamag)'


class Test1128(Test1126):
    spectrum = ('rn(spec(Zodi.fits),band(johnson,v),23.3,vegamag)+'
                '(spec(el1215a.fits)*0.2+spec(el1302a.fits)*0.01333333333+'
                'spec(el1356a.fits)*0.012+spec(el2471a.fits)*0.01)')


class Test1129(Test1126):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test1130(Test1126):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1131(CommCase):
    obsmode = 'stis,ccd,g430l,s52x2'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),23.5,vegamag)'


class Test1132(Test1131):
    spectrum = ('rn(spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits),'
                'band(johnson,v),10,vegamag)')


class Test1133(Test1131):
    spectrum = 'spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits)'


class Test1134(CommCase):
    obsmode = 'stis,ccd,g430m,c4194'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test1135(Test1134):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1136(CommCase):
    obsmode = 'stis,ccd,g430m,c4194,s52x2'
    spectrum = 'em(4300.0,1.0,1.0E-12,flam)'


class Test1137(Test1136):
    spectrum = 'spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits)'


class Test1138(CommCase):
    obsmode = 'stis,ccd,g750l,c7751'
    spectrum = ('rn(spec(Zodi.fits),band(johnson,v),22.7,vegamag)+'
                '(spec(el1215a.fits)*0.2+spec(el1302a.fits)*0.01333333333+'
                'spec(el1356a.fits)*0.012+spec(el2471a.fits)*0.01)')


class Test1139(Test1138):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.1,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1140(Test1138):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test1141(Test1138):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1142(Test1138):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),23.3,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1143(Test1138):
    spectrum = ('spec(earthshine.fits)+rn(spec(Zodi.fits),band(johnson,v),'
                '22.7,vegamag)+(spec(el1215a.fits)+spec(el1302a.fits)+'
                'spec(el1356a.fits)+spec(el2471a.fits))*2.0')


class Test1144(CommCase):
    obsmode = 'stis,ccd,g750l,c7751,s52x02'
    spectrum = 'rn(z(spec(qso_template.fits),0.03),band(johnson,v),18,vegamag)'
    force = 'extrap'


class Test1145(CommCase):
    obsmode = 'stis,ccd,g750l,c7751,s52x02'
    spectrum = 'rn(z(spec(qso_template.fits),1.0),band(johnson,v),18,vegamag)'


class Test1146(Test1145):
    spectrum = 'rn(z(spec(qso_template.fits),3.0),band(johnson,v),18,vegamag)'


class Test1147(CommCase):
    obsmode = 'stis,ccd,g750l,c7751,s52x2'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),23,vegamag)'


class Test1148(Test1147):
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),24.5,vegamag)'


class Test1149(Test1147):
    spectrum = 'spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits)'
    force = 'extrap'


class Test1150(CommCase):
    obsmode = 'stis,ccd,g750m,c7283'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1151(CommCase):
    obsmode = 'stis,ccd,g750m,c7283,s52x2'
    spectrum = 'spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits)'


class Test1152(CommCase):
    obsmode = 'stis,ccd,s03x005nd'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),4,vegamag)'


class Test1153(Test1152):
    spectrum = 'rn(unit(1.0,flam),band(johnson,v),15,vegamag)'


class Test1154(Test1152):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test1155(Test1152):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1205(CommCase):
    obsmode = 'stis,g230lb,ccd,s52x2'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test1206(Test1205):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test1207(Test1205):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test1208(CommCase):
    obsmode = 'stis,g430l,ccd,s52x2'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test1209(Test1208):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test1210(Test1208):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test1211(CommCase):
    obsmode = 'stis,g750l,ccd,s52x2'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test1212(Test1211):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test1213(Test1211):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'



"""Utility functions for commissioning tests."""

# STDLIB
import os
import sys
from collections import Iterable

# THIRD-PARTY
import numpy as np
import pytest
from numpy.testing import assert_allclose

# ASTROLIB
try:
    import pysynphot as S
    from pysynphot.spparser import parse_spec as old_parse_spec
except ImportError:
    HAS_PYSYNPHOT = False
else:
    HAS_PYSYNPHOT = True

# LOCAL
from synphot import Observation
from ..config import conf
from ..spectrum import band
from ..spparser import parse_spec

use_pysynphot = pytest.mark.skipif('not HAS_PYSYNPHOT')

# Currently, this is here because only commissioning tests are considered
# slow. If there are slow tests in the core unit tests, we can move this
# one level higher.
try:
    slow = pytest.mark.skipif(not pytest.config.getoption('--slow'),
                              reason='need --slow option to run')
except AttributeError:  # Not using pytest
    slow = pytest.mark.skipif(True, reason='need --slow option to run')

__all__ = ['use_pysynphot', 'slow', 'count_outliers', 'CommCase', 'ThermCase']


def count_outliers(data, sigma=3.0):
    """Count outliers in given data.

    This is as defined in similar method in ``SpecCase``
    in ``astrolib/pysynphot/from_commissioning/conv_base.py``.

    .. note:: This is not used but kept for reference.

    Parameters
    ----------
    data : ndarray
        Result differences to be analyzed.

    sigma : float
        Values outside this number of sigma of std. dev.
        around mean are considered outliers.

    Returns
    -------
    n_outliers : int
        Number of outlier data points.

    """
    return np.count_nonzero(abs(data) > (data.mean() + sigma * data.std()))


@use_pysynphot
@slow
@pytest.mark.remote_data
class CommCase:
    """Base class for commissioning tests."""
    obsmode = None  # Observation mode string
    spectrum = None  # SYNPHOT-like string to construct spectrum
    force = None

    # Default tables are the latest available as of 2016-07-25.
    tables = {
        'graphtable': os.path.join('mtab$OLD_FILES', '07r1502mm_tmg.fits'),
        'comptable': os.path.join('mtab$OLD_FILES', '07r1502nm_tmc.fits'),
        'thermtable': 'mtab$tae17277m_tmt.fits'}

    def setup_class(self):
        """Subclass needs to define ``obsmode`` and ``spectrum``
        class variables for this to work.

        """
        if not HAS_PYSYNPHOT:
            raise ImportError(
                'ASTROLIB PYSYNPHOT must be installed to run these tests')

        # Make sure both software use the same graph and component tables.

        conf.graphtable = self.tables['graphtable']
        conf.comptable = self.tables['comptable']
        conf.thermtable = self.tables['thermtable']

        S.setref(graphtable=self.tables['graphtable'],
                 comptable=self.tables['comptable'],
                 thermtable=self.tables['thermtable'])

        # Construct spectra for both software.

        self.sp = parse_spec(self.spectrum)
        self.bp = band(self.obsmode)

        # Astropy version has no prior knowledge of instrument-specific
        # binset, so it has to be set explicitly.
        if hasattr(self.bp, 'binset'):
            self.obs = Observation(self.sp, self.bp, force=self.force,
                                   binset=self.bp.binset)
        else:
            self.obs = Observation(self.sp, self.bp, force=self.force)

        # Astropy version does not assume a default waveset
        # (you either have it or you don't). If there is no
        # waveset, no point comparing obs waveset against ASTROLIB.
        if self.sp.waveset is None or self.bp.waveset is None:
            self._has_obswave = False
        else:
            self._has_obswave = True

        self.spref = old_parse_spec(self.spectrum)
        self.bpref = S.ObsBandpass(self.obsmode)
        self.obsref = S.Observation(self.spref, self.bpref, force=self.force)

        # Ensure we are comparing in the same units
        self.bpref.convert(self.bp._internal_wave_unit.name)
        self.spref.convert(self.sp._internal_wave_unit.name)
        self.spref.convert(self.sp._internal_flux_unit.name)
        self.obsref.convert(self.obs._internal_wave_unit.name)
        self.obsref.convert(self.obs._internal_flux_unit.name)

    @staticmethod
    def _get_new_wave(sp):
        """Astropy version does not assume a default waveset
        (you either have it or you don't). This is a convenience
        method to duck-type ASTROLIB waveset behavior.
        """
        wave = sp.waveset
        if wave is None:
            wave = conf.waveset_array
        else:
            wave = wave.value
        return wave

    def _assert_allclose(self, actual, desired, rtol=1e-07,
                         atol=sys.float_info.min):
        """``assert_allclose`` only report percentage but we
        also want to know some extra info conveniently."""
        if isinstance(actual, Iterable):
            ntot = len(actual)
        else:
            ntot = 1

        n = np.count_nonzero(
            abs(actual - desired) > atol + rtol * abs(desired))
        msg = (f'obsmode: {self.obsmode}\n'
               f'spectrum: {self.spectrum}\n'
               f'(mismatch {n}/{ntot})')
        assert_allclose(actual, desired, rtol=rtol, atol=atol, err_msg=msg)

    # TODO: Confirm whether non-default atol is acceptable.
    #       Have to use this value to avoid AssertionError for very
    #       small non-zero flux values like 1.8e-26 to 2e-311.
    def _compare_nonzero(self, new, old, thresh=0.01, atol=1e-29):
        """Compare normally when results from both are non-zero."""
        i = (new != 0) & (old != 0)

        # Make sure non-zero atol is not too high, otherwise just let it fail.
        if atol > (thresh * min(new.max(), old.max())):
            atol = sys.float_info.min

        self._assert_allclose(new[i], old[i], rtol=thresh, atol=atol)

    def _compare_zero(self, new, old, thresh=0.01):
        """Special handling for comparison when one of the results
        is zero. This is because ``rtol`` will not work."""
        i = ((new == 0) | (old == 0)) & (new != old)
        try:
            self._assert_allclose(new[i], old[i], rtol=thresh)
        except AssertionError as e:
            pytest.xfail(str(e))  # TODO: Will revisit later

    def test_band_wave(self, thresh=0.01):
        """Test bandpass waveset."""
        wave = self._get_new_wave(self.bp)
        self._assert_allclose(wave, self.bpref.wave, rtol=thresh)

    def test_spec_wave(self, thresh=0.01):
        """Test source spectrum waveset."""
        wave = self._get_new_wave(self.sp)

        # TODO: Failure due to different wavesets for blackbody; Ignore?
        try:
            self._assert_allclose(wave, self.spref.wave, rtol=thresh)
        except (AssertionError, ValueError):
            self._has_obswave = False  # Skip obs waveset tests
            if 'bb(' in self.spectrum:
                pytest.xfail('Blackbody waveset implementations are different')
            elif 'unit(' in self.spectrum:
                pytest.xfail('Flat does not use default waveset anymore')
            else:
                raise

    def test_obs_wave(self, thresh=0.01):
        """Test observation waveset."""
        if not self._has_obswave:  # Nothing to test
            return

        # Native
        wave = self.obs.waveset.value

        # TODO: Failure due to different wavesets for blackbody; Ignore?
        try:
            self._assert_allclose(wave, self.obsref.wave, rtol=thresh)
        except (AssertionError, ValueError):
            if 'bb(' in self.spectrum:
                pytest.xfail('Blackbody waveset implementations are different')
            elif 'unit(' in self.spectrum:
                self._has_obswave = False  # Skip binned flux test
                pytest.xfail('Flat does not use default waveset anymore')
            else:
                raise

        # Binned
        binset = self.obs.binset.value
        self._assert_allclose(binset, self.obsref.binwave, rtol=thresh)

    @pytest.mark.parametrize('thrutype', ['zero', 'nonzero'])
    def test_band_thru(self, thrutype, thresh=0.01):
        """Test bandpass throughput, which is always between 0 and 1."""
        wave = self.bpref.wave
        thru = self.bp(wave).value

        if thrutype == 'zero':
            self._compare_zero(thru, self.bpref.throughput, thresh=thresh)
        else:  # nonzero
            self._compare_nonzero(thru, self.bpref.throughput, thresh=thresh)

    @pytest.mark.parametrize('fluxtype', ['zero', 'nonzero'])
    def test_spec_flux(self, fluxtype, thresh=0.01):
        """Test flux for source spectrum in PHOTLAM."""
        wave = self.spref.wave
        flux = self.sp(wave).value

        if fluxtype == 'zero':
            self._compare_zero(flux, self.spref.flux, thresh=thresh)
        else:  # nonzero
            self._compare_nonzero(flux, self.spref.flux, thresh=thresh)

    @pytest.mark.parametrize('fluxtype', ['zero', 'nonzero'])
    def test_obs_flux(self, fluxtype, thresh=0.01):
        """Test flux for observation in PHOTLAM."""
        wave = self.obsref.wave
        flux = self.obs(wave).value

        # Native
        if fluxtype == 'zero':
            self._compare_zero(flux, self.obsref.flux, thresh=thresh)
        else:  # nonzero
            self._compare_nonzero(flux, self.obsref.flux, thresh=thresh)

        if not self._has_obswave:  # Do not compare binned flux
            return

        # Binned (cannot be resampled)
        binflux = self.obs.binflux.value
        if fluxtype == 'zero':
            self._compare_zero(binflux, self.obsref.binflux, thresh=thresh)
        else:  # nonzero
            try:
                self._compare_nonzero(binflux, self.obsref.binflux,
                                      thresh=thresh)
            except AssertionError as e:
                if 'unit(' in self.spectrum:
                    pytest.xfail('Flat does not use default waveset anymore:\n'
                                 f'{repr(e)}')
                else:
                    raise

    def test_countrate(self, thresh=0.01):
        """Test observation countrate calculations."""
        ans = self.obsref.countrate()

        # Astropy version does not assume a default area.
        val = self.obs.countrate(conf.area).value

        self._assert_allclose(val, ans, rtol=thresh)

    def test_efflam(self, thresh=0.01):
        """Test observation effective wavelength."""
        ans = self.obsref.efflam()
        val = self.obs.effective_wavelength().value
        self._assert_allclose(val, ans, rtol=thresh)

    def teardown_class(self):
        """Reset config for both software."""
        for cfgname in self.tables:
            conf.reset(cfgname)

        S.setref()


class ThermCase(CommCase):
    """Commissioning tests with thermal component."""

    @pytest.mark.parametrize('fluxtype', ['zero', 'nonzero'])
    def test_therm_spec(self, fluxtype, thresh=0.01):
        """Test bandpass thermal spectrum."""
        thspref = self.bpref.obsmode.ThermalSpectrum()
        thsp = self.bp.obsmode.thermal_spectrum()

        # Make sure comparing same units
        thspref.convert(thsp._internal_wave_unit.name)
        thspref.convert(thsp._internal_flux_unit.name)

        # waveset not expected to be same here, so just compare flux
        flux = thsp(thspref.wave).value
        if fluxtype == 'zero':
            self._compare_zero(flux, thspref.flux, thresh=thresh)
        else:  # nonzero
            # TODO: Is the refactored version really better?
            try:
                self._compare_nonzero(flux, thspref.flux, thresh=thresh)
            except AssertionError:
                pytest.xfail('New thermal spectrum samples better')

    def test_thermback(self, thresh=0.01):
        """Test bandpass thermal background."""
        ans = self.bpref.thermback()
        val = self.bp.thermback().value

        self._assert_allclose(val, ans, rtol=thresh)



"""This module contains WFC3/IR commissioning tests.
Adapted from ``astrolib/pysynphot/from_commissioning/wfc3_ir/test*.py``.
"""

# STDLIB
import os
import shutil

# THIRD-PARTY
from astropy.utils.data import get_pkg_data_filename

# LOCAL
from ..utils import ThermCase

# Local test data
datafiles = ['earthshine.fits', 'el1215a.fits', 'el1302a.fits', 'el1356a.fits',
             'el2471a.fits', 'Zodi.fits']


def setup_module(module):
    """Copy test data to working directory so ``parse_spec`` works
    properly for both ``stsynphot`` and ASTROLIB PYSYNPHOT."""
    for datafile in datafiles:
        src = get_pkg_data_filename(os.path.join('data', datafile))
        shutil.copyfile(src, datafile)


def teardown_module(module):
    """Clean up test data in working directory."""
    for datafile in datafiles:
        os.remove(datafile)


class Test1361(ThermCase):
    # Original test used gal1 but it is no longer supported, so we use gal3
    obsmode = 'wfc3,ir,f160w'  # IRAF thermback=0.1359
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_100.fits)'
                ',band(cousins,i),28.0,vegamag)*ebmvx(0.04,gal3)')


class Test1362(Test1361):
    # Original test used smc but it is no longer supported, so we use smcbar
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_100.fits)'
                ',band(cousins,i),28.0,vegamag)*ebmvx(0.16,smcbar)')


class Test1363(Test1361):
    # Original test used lmc but it is no longer supported, so we use lmcavg
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_100.fits)'
                ',band(cousins,i),28.0,vegamag)*ebmvx(0.2,lmcavg)')


class Test1364(Test1361):
    # Original test used xgal but it is no longer supported, so we use xgalsb
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_100.fits)'
                ',band(cousins,i),28.0,vegamag)*ebmvx(0.24,xgalsb)')


class Test1365(Test1361):
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_11.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.12,gal3)')


class Test1366(Test1361):
    # Original test used lmc but it is no longer supported, so we use lmcavg
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_11.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.2,lmcavg)')


class Test1367(Test1361):
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_114.fits)'
                ',band(cousins,i),28.0,vegamag)*ebmvx(0.12,gal3)')


class Test1368(Test1361):
    # Original test used smc but it is no longer supported, so we use smcbar
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_117.fits)'
                ',band(cousins,i),28.0,vegamag)*ebmvx(0.16,smcbar)')


class Test1369(Test1361):
    # Original test used lmc but it is no longer supported, so we use lmcavg
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_118.fits)'
                ',band(cousins,i),28.0,vegamag)*ebmvx(0.2,lmcavg)')


class Test1370(Test1361):
    # Original test used smc but it is no longer supported, so we use smcbar
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_12.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.16,smcbar)')


class Test1371(Test1361):
    # Original test used xgal but it is no longer supported, so we use xgalsb
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_12.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.24,xgalsb)')


class Test1372(Test1361):
    # Original test used lmc but it is no longer supported, so we use lmcavg
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_13.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.2,lmcavg)')


class Test1373(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_14.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.08,gal3)')


class Test1374(Test1361):
    # Original test used xgal but it is no longer supported, so we use xgalsb
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_14.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.24,xgalsb)')


class Test1375(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_15.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.04,gal3)')


class Test1376(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_16.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.08,gal3)')


class Test1377(Test1361):
    # Original test used smc but it is no longer supported, so we use smcbar
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_16.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.16,smcbar)')


class Test1378(Test1361):
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_17.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.12,gal3)')


class Test1379(Test1361):
    # Original test used lmc but it is no longer supported, so we use lmcavg
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_17.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.2,lmcavg)')


class Test1380(Test1361):
    # Original test used smc but it is no longer supported, so we use smcbar
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_18.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.16,smcbar)')


class Test1381(Test1361):
    # Original test used xgal but it is no longer supported, so we use xgalsb
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_18.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.24,xgalsb)')


class Test1382(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_19.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.04,gal3)')


class Test1383(Test1361):
    # Original test used lmc but it is no longer supported, so we use lmcavg
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_19.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.2,lmcavg)')


class Test1384(Test1361):
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_2.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.12,gal3)')


class Test1385(Test1361):
    # Original test used smc but it is no longer supported, so we use smcbar
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_2.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.16,smcbar)')


class Test1386(Test1361):
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_20.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.12,gal3)')


class Test1387(Test1361):
    # Original test used xgal but it is no longer supported, so we use xgalsb
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_20.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.24,xgalsb)')


class Test1388(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_22.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.08,gal3)')


class Test1389(Test1361):
    # Original test used xgal but it is no longer supported, so we use xgalsb
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_23.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.24,xgalsb)')


class Test1390(Test1361):
    # Original test used smc but it is no longer supported, so we use smcbar
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_24.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.16,smcbar)')


class Test1391(Test1361):
    # Original test used lmc but it is no longer supported, so we use lmcavg
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_25.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.2,lmcavg)')


class Test1392(Test1361):
    # Original test used xgal but it is no longer supported, so we use xgalsb
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_26.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.24,xgalsb)')


class Test1393(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_27.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.04,gal3)')


class Test1394(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_29.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.08,gal3)')


class Test1395(Test1361):
    # Original test used lmc but it is no longer supported, so we use lmcavg
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_3.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.2,lmcavg)')


class Test1396(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_31.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.04,gal3)')


class Test1397(Test1361):
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_33.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.12,gal3)')


class Test1398(Test1361):
    # Original test used smc but it is no longer supported, so we use smcbar
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_34.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.16,smcbar)')


class Test1399(Test1361):
    # Original test used xgal but it is no longer supported, so we use xgalsb
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_36.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.24,xgalsb)')


class Test1400(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_37.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.04,gal3)')


class Test1401(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_38.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.08,gal3)')


class Test1402(Test1361):
    # Original test used xgal but it is no longer supported, so we use xgalsb
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_4.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.24,xgalsb)')


class Test1403(Test1361):
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_40.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.12,gal3)')


class Test1404(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_5.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.04,gal3)')


class Test1405(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_5.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.08,gal3)')


class Test1406(Test1361):
    # Original test used xgal but it is no longer supported, so we use xgalsb
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_50.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.24,xgalsb)')


class Test1407(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_51.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.04,gal3)')


class Test1408(Test1361):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_52.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.08,gal3)')


class Test1409(Test1361):
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_53.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.12,gal3)')


class Test1410(Test1361):
    # Original test used smc but it is no longer supported, so we use smcbar
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_54.fits),'
                'band(cousins,i),28.0,vegamag)*ebmvx(0.16,smcbar)')


class Test1461(ThermCase):
    obsmode = 'wfc3,ir,g141'
    spectrum = 'rn(icat(k93models,9230,0.0,4.1),band(johnson,v),18.0,vegamag)'


class Test1462(Test1461):
    spectrum = 'rn(icat(k93models,9230,0.0,4.1),band(johnson,v),23.0,vegamag)'


class Test1463(Test1461):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_1.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.04,gal3)')


class Test1464(Test1461):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_1.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.08,gal3)')


class Test1465(Test1461):
    # Original test used smc but it is no longer supported, so we use smcbar
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_10.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.16,smcbar)')


class Test1466(Test1461):
    # Original test used lmc but it is no longer supported, so we use lmcavg
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_11.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.2,lmcavg)')


class Test1467(Test1461):
    # Original test used xgal but it is no longer supported, so we use xgalsb
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_12.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.24,xgalsb)')


class Test1468(Test1461):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_14.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.08,gal3)')


class Test1469(Test1461):
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_2.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.12,gal3)')


class Test1470(Test1461):
    # Original test used smc but it is no longer supported, so we use smcbar
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_2.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.16,smcbar)')


class Test1471(Test1461):
    # Original test used lmc but it is no longer supported, so we use lmcavg
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_3.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.2,lmcavg)')


class Test1472(Test1461):
    # Original test used xgal but it is no longer supported, so we use xgalsb
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_4.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.24,xgalsb)')


class Test1473(Test1461):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_5.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.04,gal3)')


class Test1474(Test1461):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_5.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.08,gal3)')


class Test1475(Test1461):
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_6.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.12,gal3)')


class Test1476(Test1461):
    # Original test used gal1 but it is no longer supported, so we use gal3
    spectrum = ('rn(spec($PYSYN_CDBS/grid/pickles/dat_uvk/pickles_uk_9.fits),'
                'band(cousins,i),23.0,vegamag)*ebmvx(0.04,gal3)')


class Test1477(ThermCase):
    obsmode = 'wfc3,ir,g141,bkg'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),band(johnson,v),'
                '21.7,vegamag)+(spec(el1215a.fits)+spec(el1302a.fits)+'
                'spec(el1356a.fits)+spec(el2471a.fits))')


class Test1478(Test1477):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),band(johnson,v),'
                '22.1,vegamag)+(spec(el1215a.fits)+spec(el1302a.fits)+'
                'spec(el1356a.fits)+spec(el2471a.fits))')


class Test1479(Test1477):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),band(johnson,v),'
                '22.424602593467696,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1480(Test1477):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),band(johnson,v),'
                '22.7,vegamag)+(spec(el1215a.fits)+spec(el1302a.fits)+'
                'spec(el1356a.fits)+spec(el2471a.fits))')


class Test1481(Test1477):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),band(johnson,v),'
                '23.3,vegamag)+(spec(el1215a.fits)+spec(el1302a.fits)+'
                'spec(el1356a.fits)+spec(el2471a.fits))')


class Test1482(Test1477):
    spectrum = ('spec(earthshine.fits)*0.5+spec(Zodi.fits)*0.5+'
                '(spec(el1215a.fits)+spec(el1302a.fits)+spec(el1356a.fits)+'
                'spec(el2471a.fits))')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Module to handle observations based on observation modes."""

# STDLIB
import re
import warnings

# THIRD-PARTY
import numpy as np

# ASTROPY
from astropy import log
from astropy import units as u
from astropy.utils.decorators import lazyproperty
from astropy.utils.exceptions import AstropyUserWarning

# SYNPHOT
from synphot import exceptions as synexceptions
from synphot import units
from synphot.models import Empirical1D
from synphot.spectrum import SourceSpectrum, SpectralElement
from synphot.thermal import ThermalSpectralElement
from synphot.utils import merge_wavelengths

# LOCAL
from . import exceptions, stio
from .config import conf
from .spectrum import Vega, interpolate_spectral_element
from .tables import GraphTable, CompTable
from .wavetable import WAVECAT

__all__ = ['reset_cache', 'Component', 'ThermalComponent',
           'BaseObservationMode', 'ObservationMode', 'ThermalObservationMode']

_band_patt = re.compile(r'band\((.*?)\)', re.IGNORECASE)

# Cache previously loaded graph, component, thermal, and detector tables
_GRAPHDICT = {}
_COMPDICT = {}
_THERMDICT = {}
_DETECTORDICT = {}


def reset_cache():
    """Empty the table dictionaries cache."""
    global _GRAPHDICT, _COMPDICT, _THERMDICT, _DETECTORDICT
    _GRAPHDICT.clear()
    _COMPDICT.clear()
    _THERMDICT.clear()
    _DETECTORDICT.clear()


class Component:
    """Class to handle individual components in `BaseObservationMode`.

    Parameters
    ----------
    throughput_name : str
        Component filename.

    interpval : float or `None`
        If not `None`, interpolate to the given value.
        See :func:`stsynphot.spectrum.interpolate_spectral_element`.

    Attributes
    ----------
    throughput_name : str
        Component filename.

    throughput : `synphot.spectrum.SpectralElement` or `None`
        Component spectrum object.

    """
    def __init__(self, throughput_name, interpval=None):
        self.throughput_name = throughput_name

        # Extract bandpass unless component is a CLEAR filter.
        if throughput_name != conf.clear_filter:
            if interpval is None:
                self.throughput = SpectralElement.from_file(throughput_name)
            else:
                self.throughput = interpolate_spectral_element(
                    throughput_name, interpval)
        else:
            self.throughput = None

    @property
    def empty(self):
        """`True` if ``self.throughput`` is empty."""
        return self.throughput is None

    def __str__(self):
        return str(self.throughput)


class ThermalComponent(Component):
    """Class to handle thermal components in `BaseObservationMode`.

    Parameters
    ----------
    throughput_name, thermal_name : str
        Optical and thermal component filenames.

    interpval : float or `None`
        If not `None`, interpolate to the given value (optical only).
        See :func:`stsynphot.spectrum.interpolate_spectral_element`.


    Attributes
    ----------
    throughput_name, thermal_name : str
        Optical and thermal component filenames.

    throughput : `synphot.spectrum.SpectralElement` or `None`
        Optical component spectrum object.

    emissivity : `synphot.thermal.ThermalSpectralElement` or `None`
        Thermal component spectrum object.

    """
    def __init__(self, throughput_name, thermal_name, interpval=None):
        super(ThermalComponent, self).__init__(
            throughput_name, interpval=interpval)
        self.thermal_name = thermal_name

        # Extract thermal spectrum unless component is a CLEAR filter.
        if thermal_name != conf.clear_filter:
            self.emissivity = ThermalSpectralElement.from_file(thermal_name)
        else:
            self.emissivity = None

    @property
    def empty(self):
        """`True` if component is empty."""
        return self.throughput is None and self.emissivity is None

    def __str__(self):
        return str(self.emissivity)


def _process_graphtable(graphtable):
    """Load and cache graphtable. Get primary area."""
    global _GRAPHDICT

    # Use default graph table if not given
    if graphtable is None:
        gtname = conf.graphtable
    else:
        gtname = graphtable

    # Load and cache graphtable, if not in cache
    if gtname in _GRAPHDICT:
        gt = _GRAPHDICT[gtname]
    else:
        gt = GraphTable(gtname)
        _GRAPHDICT[gtname] = gt

    # Set telescope collecting area
    if gt.primary_area is None:
        area = conf.area
    else:
        area = gt.primary_area
    primary_area = units.validate_quantity(area, units.AREA)

    return gt, gtname, primary_area


class BaseObservationMode:
    """Base class to handle an observation that uses the graph
    and optical component tables, common to both optical and
    thermal observation modes.

    .. note::

        ``modes`` for parameterized mode is set such that the parameterized
        value is stripped away; e.g., ``'acs,wfc1,f555w,mjd#53000'``
        becomes ``['acs', 'wfc1', 'f555w', 'mjd#']``. Instead, the
        parameterized value is kept in ``pardict``; e.g., ``'mjd#53000'``
        becomes ``{'mjd': 53000.0}``.

        ``pixscale`` is set from ``stsynphot.config.conf.detectorfile``,
        which is parsed with :func:`stsynphot.stio.read_detector_pars`.

        ``binset`` is set by ``stsynphot.wavetable.WAVECAT``.

    Parameters
    ----------
    obsmode : str
        Observation mode.

    graphtable : str or `None`
        Graph table filename.
        If `None`, uses ``stsynphot.config.conf.graphtable``.

    comptable : str or `None`
        Optical component table filename.
        If `None`, uses ``stsynphot.config.conf.comptable``.

    Attributes
    ----------
    modes : list
        List of individual modes within observation mode.

    pardict : dict
        Maps parameterized mode to its value.

    gtname, ctname : str
        Graph and component table filenames.

    compnames, thcompnames : list of str
        Optical and thermal components.

    components : list of obj
        List of component objects.

    primary_area : `astropy.units.quantity.Quantity`
        Telescope collecting area.

    pixscale : `astropy.units.quantity.Quantity`
        Detector pixel scale.

    binset : str
        Wavelength table filename/param string from matching obsmode.

    bandwave : `astropy.units.quantity.Quantity`
        Wavelength set defined by ``binset``.

    """
    def __init__(self, obsmode, graphtable=None, comptable=None):
        global _GRAPHDICT, _COMPDICT

        # Strip "band()" syntax if present, and force lowercase
        tmatch = _band_patt.search(obsmode)
        if tmatch:
            self._obsmode = tmatch.group(1).lower()
        else:
            self._obsmode = obsmode.lower()

        # Split obsmode and separate parameterized modes
        modes = self._obsmode.replace(' ', '').split(',')
        self.pardict = {}
        if '#' in self._obsmode:
            self.modes = []
            for m in modes:
                if '#' in m:
                    key, val = m.split('#')
                    self.pardict[key] = float(val)
                    self.modes.append(f'{key:s}#')
                else:
                    self.modes.append(m)
        else:
            self.modes = modes

        # Get graph table and primary area
        gt, self.gtname, self.primary_area = _process_graphtable(graphtable)

        # Get optical and thermal components
        self.compnames, self.thcompnames = gt.get_comp_from_gt(self.modes, 1)

        # Use default optical component table if not given
        if comptable is None:
            self.ctname = conf.comptable
        else:
            self.ctname = comptable

        # Load and cache comptable, if not in cache
        if self.ctname in _COMPDICT:
            ct = _COMPDICT[self.ctname]
        else:
            ct = CompTable(self.ctname)
            _COMPDICT[self.ctname] = ct

        # Set by sub-classes
        self.components = None

        # Get optical component filenames
        self._throughput_filenames = ct.get_filenames(self.compnames)

        # Set detector pixel scale
        self._set_pixscale()

        # For sensitivity calculations
        self._constant = self.primary_area / units.HC

        # Get wavelength set
        try:
            self.binset, self.bandwave = WAVECAT.load_waveset(self._obsmode)
        except (KeyError, exceptions.AmbiguousObsmode):
            self.binset = ''
            self.bandwave = None

    def _set_pixscale(self):
        """Set pixel scale.
        If multiple matches found, only first match is used.

        """
        global _DETECTORDICT

        det_file = stio.irafconvert(conf.detectorfile)

        if det_file in _DETECTORDICT:
            data = _DETECTORDICT[det_file]
        else:
            data = stio.read_detector_pars(det_file)
            _DETECTORDICT[det_file] = data

        obsmode = ','.join(self._obsmode.split(',')[:2])
        pixscales = data[data['OBSMODE'] == obsmode]['SCALE'].data

        if pixscales.size < 1:
            self.pixscale = None
        else:
            self.pixscale = pixscales[0] * u.arcsec

    def _get_components(self):
        raise NotImplementedError('To be implemented by subclasses.')

    def __str__(self):
        return self._obsmode

    def __len__(self):
        return len(self.components)

    @staticmethod
    def _parkey_from_filename(filename):
        """Extract parkey from component filename in the format of
        ``name[parkey#]``.

        """
        if filename.endswith('#]'):
            x = filename.split('[')
            parkey = x[1][:-2]
        else:
            parkey = None
        return parkey

    def showfiles(self):  # pragma: no cover
        """Display optical component filenames.

        .. note:: Similar to IRAF SYNPHOT SHOWFILES.

        """
        info_str = '#Throughput table names:\n'
        for name in self._throughput_filenames:
            if name != conf.clear_filter:
                info_str += f'{name}\n'
        log.info(info_str.rstrip())


class ObservationMode(BaseObservationMode):
    """Class to handle an observation that uses the graph
    and optical component tables.

    See `BaseObservationMode` for additional attributes.

    Parameters
    ----------
    obsmode, graphtable, comptable
        See `BaseObservationMode`.

    component_dict : dict
        Maps component filename to corresponding `Component`.

    """
    def __init__(self, obsmode, graphtable=None, comptable=None,
                 component_dict={}):
        super(ObservationMode, self).__init__(
            obsmode, graphtable=graphtable, comptable=comptable)
        self._component_dict = component_dict
        self.components = self._get_components()

    def _get_components(self):
        """Get optical components."""
        components = []

        for throughput_name in self._throughput_filenames:
            parkey = self._parkey_from_filename(throughput_name)

            # Hack for WFPC2 ramp filters
            if parkey == 'wave' and self.modes[0] == 'wfpc2':
                parkey = 'lrf'

            cdict_key = (throughput_name, self.pardict.get(parkey))

            if cdict_key not in self._component_dict:
                self._component_dict[cdict_key] = Component(
                    cdict_key[0], interpval=cdict_key[1])

            component = self._component_dict[cdict_key]

            if not component.empty:
                components.append(component)

        return components

    def _mul_thru(self, index):
        """Multiply all component spectra starting at given index."""
        product = self.components[index].throughput

        if len(self.components) > index:
            for component in self.components[index + 1:]:
                if not component.empty:
                    product = product * component.throughput

        product.meta['header'] = ''  # Clean up messy header
        return product

    @lazyproperty
    def throughput(self):
        """Combined throughput from multiplying all the components together."""
        try:
            thru = self._mul_thru(0)
        except IndexError as e:  # pragma: no cover
            thru = None
            warnings.warn(
                f'Graph table is broken: {repr(e)}', AstropyUserWarning)
        return thru

    @lazyproperty
    def sensitivity(self):
        """Sensitivity spectrum to convert flux in
        :math:`erg \\; cm^{-2} \\; s^{-1} \\; \\AA^{-1}` to
        :math:`count s^{-1} \\AA^{-1}`. Calculation is done by
        combining the throughput curves with
        :math:`\\frac{h \\; c}{\\lambda}` .

        """
        x = self.throughput.waveset
        y = self.throughput(x)
        thru = y.value * x.value * self._constant.value
        meta = {'expr': f'Sensitivity for {self._obsmode}'}
        return SpectralElement(
            Empirical1D, points=x, lookup_table=thru, meta=meta)

    def thermal_spectrum(self, thermtable=None):
        """Calculate thermal spectrum using
        :func:`ThermalObservationMode.to_spectrum`.

        Parameters
        ----------
        thermtable : str or `None`
            Thermal component table filename.
            If `None`, uses ``stsynphot.config.conf.thermtable``.

        Returns
        -------
        sp : `synphot.spectrum.SourceSpectrum`
            Thermal spectrum in PHOTLAM.

        Raises
        ------
        synphot.exceptions.SynphotError
            Calculation failed.

        """
        thom = ThermalObservationMode(
            self._obsmode, graphtable=self.gtname, comptable=self.ctname,
            thermtable=thermtable)

        try:
            sp = thom.to_spectrum()
        except IndexError as e:  # pragma: no cover
            raise synexceptions.SynphotError(f'Broken graph table: {repr(e)}')

        return sp


class ThermalObservationMode(BaseObservationMode):
    """Class to handle an observation that uses the graph,
    optical component, and thermal component tables.

    See `BaseObservationMode` for additional attributes.

    Parameters
    ----------
    obsmode, graphtable, comptable
        See `BaseObservationMode`.

    thermtable : str or `None`
        Thermal component table filename.
        If `None`, uses ``stsynphot.config.conf.thermtable``.

    Attributes
    ----------
    thname : str
        Thermal component table filename.

    Raises
    ------
    NotImplementedError
        Clear filters only not supported.

    """
    def __init__(self, obsmode, graphtable=None, comptable=None,
                 thermtable=None):
        global _GRAPHDICT, _COMPDICT, _THERMDICT

        super(ThermalObservationMode, self).__init__(
            obsmode, graphtable=graphtable, comptable=comptable)

        # Check here to see if there are any valid filters.
        # "0.0" was added in tae17277m_tmt.fits (Apr 2017).
        if set(self.thcompnames).issubset(set([conf.clear_filter, '', '0.0'])):
            raise NotImplementedError(
                f'No thermal support provided for {self._obsmode}')

        # Use default thermal component table if not given
        if thermtable is None:
            self.thname = conf.thermtable
        else:
            self.thname = thermtable

        # Load and cache thermtable, if not in cache
        if self.thname in _THERMDICT:
            thct = _THERMDICT[self.thname]
        else:
            thct = CompTable(self.thname)
            _THERMDICT[self.thname] = thct

        # Get thermal component filenames
        self._thermal_filenames = thct.get_filenames(self.thcompnames)

        self.components = self._get_components()

    def _get_components(self):
        """Get thermal components."""
        components = []

        for throughput_name, thermal_name in zip(
                self._throughput_filenames, self._thermal_filenames):
            parkey = self._parkey_from_filename(throughput_name)
            component = ThermalComponent(throughput_name, thermal_name,
                                         interpval=self.pardict.get(parkey))
            if not component.empty:
                components.append(component)

        return components

    def __str__(self):
        return f'{self._obsmode} (thermal)'

    def _merge_em_wave(self):
        """Merge emissivity wavelength sets in Angstrom."""
        index = 1

        # Find first wavelength set
        for component in self.components:
            emissivity = component.emissivity
            if emissivity is None:
                index += 1
            else:
                result = emissivity.waveset.value
                break

        # Merge subsequent wavelength sets
        for component in self.components[index:]:
            emissivity = component.emissivity
            if emissivity is not None:
                result = merge_wavelengths(result, emissivity.waveset.value)

        return result

    def _get_wave_intersection(self):
        """Find wavelengths in Angstrom where
        ``stsynphot.config.conf.waveset_array`` intersects with
        ``stsynphot.spectrum.Vega``.

        .. note:: No one knows why Vega is the chosen one.

        """
        def_wave = conf.waveset_array  # Angstrom
        minw = min(def_wave)
        maxw = max(def_wave)

        # Refine min and max wavelengths using thermal components
        for component in self.components[1:]:
            emissivity = component.emissivity
            if emissivity is not None:
                w = emissivity.waveset.value  # Angstrom
                minw = max(minw, w.min())
                maxw = min(maxw, w.max())

        w = self._merge_em_wave()
        result = w[(w > minw) & (w < maxw)]

        # Intersect with Vega
        if Vega is None:
            raise synexceptions.SynphotError('Missing Vega spectrum.')
        w = Vega.waveset.value  # Angstrom

        return result[(result > w.min()) & (result < w.max())]

    def to_spectrum(self):
        """Get thermal spectrum.

        The calculations start with zero-flux spectrum.
        Then for each component:

        #. Multiply with optical throughput.
        #. Add in thermal source spectrum from
           :meth:`synphot.thermal.ThermalSpectralElement.thermal_source`.

        Returns
        -------
        sp : `synphot.spectrum.SourceSpectrum`
            Thermal spectrum in PHOTLAM.

        """
        # Create zero-flux spectrum
        x = self._get_wave_intersection()  # Angstrom
        y = np.zeros_like(x, dtype=np.float64)  # PHOTLAM
        minw, maxw = x[([0, -1], )]
        sp = SourceSpectrum(Empirical1D, points=x, lookup_table=y)

        for component in self.components:
            # Transmissive section (optical passband)
            if component.throughput is not None:
                sp = sp * component.throughput

            # Thermal section
            if component.emissivity is not None:
                sp = sp + component.emissivity.thermal_source()

                # Trim spectrum
                w = sp.waveset.value
                mask = (w >= minw) & (w <= maxw)
                x = w[mask]
                y = sp(x)
                sp = SourceSpectrum(Empirical1D, points=x, lookup_table=y)

        meta = {'expr': f'{self._obsmode} ThermalSpectrum'}
        sp.meta.update(meta)
        return sp



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test spparser.py module, which uses spark.py.

.. note::

    Only testing to see if the parser makes the right kind of
    objects. Quality of the data is tested in other modules.

"""

# STDLIB
import os

# THIRD-PARTY
import pytest
from astropy import units as u
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.exceptions import AstropyUserWarning
from numpy.testing import assert_allclose

# SYNPHOT
from synphot import exceptions as synexceptions
from synphot import units
from synphot.models import (BlackBodyNorm1D, Box1D, ConstFlux1D, Empirical1D,
                            GaussianFlux1D, PowerLawFlux1D)
from synphot.reddening import ExtinctionCurve
from synphot.spectrum import SourceSpectrum, SpectralElement

# LOCAL
from .. import catalog, exceptions, observationmode, spectrum, spparser
from ..config import conf
from ..stio import resolve_filename


def _single_functioncall(sp, ans_cls, ans_model, ans_name, ans_z=0):
    assert isinstance(sp, ans_cls)

    # Do not check composite model
    if ans_model is not None:
        assert isinstance(sp.model, ans_model)

    if ans_name:
        assert sp.meta['expr'] == ans_name
    if ans_z is not None:
        assert_allclose(sp.z, ans_z)


def _compare_spectra(sp1, sp2):
    """Test that two spectra are basically equivalent."""
    if sp1.waveset is None:
        assert sp2.waveset is None
        w = [100, 5000, 11000] * u.AA
    else:
        w = sp1.waveset
        assert_quantity_allclose(w, sp2.waveset)
    assert_quantity_allclose(sp1(w), sp2(w))
    assert_quantity_allclose(sp1.integrate(wavelengths=w),
                             sp2.integrate(wavelengths=w))
    assert type(sp1.model.__class__) == type(sp2.model.__class__)  # noqa
    if hasattr(sp1, 'z'):
        assert sp1.z == sp2.z


def test_unit_1_flam():
    sp1 = spparser.parse_spec('unit(1, flam)')
    _single_functioncall(sp1, SourceSpectrum, ConstFlux1D, 'unit(1.0,flam)')

    sp2 = SourceSpectrum(ConstFlux1D, amplitude=1 * units.FLAM)
    _compare_spectra(sp1, sp2)


def test_bb_5000():
    sp1 = spparser.parse_spec('bb(5000)')
    _single_functioncall(sp1, SourceSpectrum, BlackBodyNorm1D, 'bb(5000.0)')

    sp2 = SourceSpectrum(BlackBodyNorm1D, temperature=5000 * u.K)
    _compare_spectra(sp1, sp2)


def test_powerlaw_5000_1_flam():
    sp1 = spparser.parse_spec('pl(5000, 1, flam)')
    _single_functioncall(
        sp1, SourceSpectrum, PowerLawFlux1D, 'pl(5000.0,1.0,flam)')

    sp2 = SourceSpectrum(PowerLawFlux1D, amplitude=1 * units.FLAM,
                         x_0=5000 * u.AA, alpha=-1)
    _compare_spectra(sp1, sp2)


def test_box_5000_1():
    sp1 = spparser.parse_spec('box(5000, 1)')
    _single_functioncall(sp1, SpectralElement, Box1D, 'box(5000.0,1.0)',
                         ans_z=None)

    sp2 = SpectralElement(Box1D, amplitude=1, x_0=5000 * u.AA, width=1 * u.AA)
    _compare_spectra(sp1, sp2)


def test_em_5000_25_1_flam():
    sp1 = spparser.parse_spec('em(5000, 25, 1, flam)')
    _single_functioncall(
        sp1, SourceSpectrum, GaussianFlux1D, 'em(5000, 25, 1, FLAM)')

    f = 1 * (units.FLAM * u.AA)  # Integrated flux
    sp2 = SourceSpectrum(
        GaussianFlux1D, mean=5000 * u.AA, fwhm=25 * u.AA, total_flux=f)
    _compare_spectra(sp1, sp2)


def test_rn_bb_box_abmag():
    sp1 = spparser.parse_spec('rn(bb(5000), box(5000, 10), 17, abmag)')
    _single_functioncall(sp1, SourceSpectrum, None,
                         'rn(bb(5000.0),box(5000.0,10.0),17.0,abmag)')

    bb = SourceSpectrum(BlackBodyNorm1D, temperature=5000 * u.K)
    box = SpectralElement(Box1D, amplitude=1, x_0=5000 * u.AA, width=10 * u.AA)
    sp2 = bb.normalize(17 * u.ABmag, band=box)
    _compare_spectra(sp1, sp2)


def test_z_null():
    """ETC junk spectrum results in flat spectrum with no redshift."""
    sp1 = spparser.parse_spec('z(null, 0.1)')
    _single_functioncall(sp1, SourceSpectrum, ConstFlux1D, 'z(null,0.1)')

    sp2 = SourceSpectrum(ConstFlux1D, amplitude=1 * units.PHOTLAM)
    _compare_spectra(sp1, sp2)


def test_z_em():
    sp1 = spparser.parse_spec('z(em(5000, 25, 1, flam), 0.1)')
    _single_functioncall(
        sp1, SourceSpectrum, None, 'z(em(5000, 25, 1, FLAM),0.1)', ans_z=0.1)

    f = 1 * (units.FLAM * u.AA)  # Integrated flux
    sp2 = SourceSpectrum(
        GaussianFlux1D, mean=5000 * u.AA, fwhm=25 * u.AA, total_flux=f)
    sp2.z = 0.1
    _compare_spectra(sp1, sp2)


@pytest.mark.remote_data
def test_remote_spec_vegafile():
    sp1 = spparser.parse_spec('spec(crcalspec$alpha_lyr_stis_007.fits)')
    _single_functioncall(sp1, SourceSpectrum, Empirical1D,
                         'spec(crcalspec$alpha_lyr_stis_007.fits)')

    sp2 = SourceSpectrum.from_file(resolve_filename(
        os.environ['PYSYN_CDBS'], 'calspec', 'alpha_lyr_stis_007.fits'))
    _compare_spectra(sp1, sp2)


@pytest.mark.remote_data
def test_remote_band_v():
    sp1 = spparser.parse_spec('band(v)')
    _single_functioncall(
        sp1, spectrum.ObservationSpectralElement, Empirical1D, 'band(v)',
        ans_z=None)

    sp2 = SpectralElement.from_filter('johnson_v')
    _compare_spectra(sp1, sp2)


@pytest.mark.remote_data
def test_remote_icat_k93():
    sp1 = spparser.parse_spec('icat(k93models, 5000, 0.5, 0)')
    _single_functioncall(sp1, SourceSpectrum, Empirical1D,
                         'k93models(T_eff=5000,metallicity=0.5,log_g=0)')

    sp2 = catalog.grid_to_spec('k93models', 5000, 0.5, 0)
    _compare_spectra(sp1, sp2)


@pytest.mark.remote_data
def test_remote_ebmvx_mwavg():
    sp1 = spparser.parse_spec('ebmvx(0.3, mwavg)')
    _single_functioncall(
        sp1, ExtinctionCurve, Empirical1D, 'ebmvx(0.3,mwavg)', ans_z=None)

    sp2 = spectrum.ebmvx('mwavg', 0.3)
    _compare_spectra(sp1, sp2)


@pytest.mark.remote_data
def test_remote_rn_calspec_box():
    sp1 = spparser.parse_spec(
        'rn(crcalspec$gd71_mod_005.fits, box(5000, 10), 17, vegamag)')
    _single_functioncall(
        sp1, SourceSpectrum, None,
        'rn(crcalspec$gd71_mod_005.fits,box(5000.0,10.0),17.0,vegamag)')

    gd71 = SourceSpectrum.from_file(resolve_filename(
        os.environ['PYSYN_CDBS'], 'calspec', 'gd71_mod_005.fits'))
    box = SpectralElement(Box1D, amplitude=1, x_0=5000 * u.AA, width=10 * u.AA)
    sp2 = gd71.normalize(17 * units.VEGAMAG, band=box, vegaspec=spectrum.Vega)
    _compare_spectra(sp1, sp2)


@pytest.mark.remote_data
def test_remote_rn_icat_k93():
    sp1 = spparser.parse_spec(
        'rn(icat(k93models, 5000, 0.5, 0), '
        'cracscomp$acs_f814w_hrc_006_syn.fits, 17, obmag)')
    _single_functioncall(
        sp1, SourceSpectrum, None,
        'rn(k93models(T_eff=5000,metallicity=0.5,log_g=0),'
        'cracscomp$acs_f814w_hrc_006_syn.fits,17.0,obmag)')

    k93 = catalog.grid_to_spec('k93models', 5000, 0.5, 0)
    bp = SpectralElement.from_file(resolve_filename(
        os.environ['PYSYN_CDBS'], 'comp', 'acs', 'acs_f814w_hrc_006_syn.fits'))
    sp2 = k93.normalize(17 * units.OBMAG, band=bp, area=conf.area)
    _compare_spectra(sp1, sp2)


@pytest.mark.remote_data
def test_remote_rn_powerlaw():
    sp1 = spparser.parse_spec('rn(pl(5000, 1, flam), band(v), 1, photlam)')
    _single_functioncall(sp1, SourceSpectrum, None,
                         'rn(pl(5000.0,1.0,flam),band(v),1.0,photlam)')

    pl = SourceSpectrum(PowerLawFlux1D, amplitude=1 * units.FLAM,
                        x_0=5000 * u.AA, alpha=-1)
    bp = SpectralElement.from_filter('johnson_v')
    sp2 = pl.normalize(1 * units.PHOTLAM, band=bp)
    _compare_spectra(sp1, sp2)


@pytest.mark.remote_data
def test_remote_rn_unit_1_flam():
    sp1 = spparser.parse_spec(
        'rn(unit(1,flam), band(acs, wfc1, fr388n#3881.0), 10, abmag)')
    _single_functioncall(
        sp1, SourceSpectrum, None,
        'rn(unit(1.0,flam),band(acs,wfc1,fr388n#3881.0),10.0,abmag)')

    constsp = SourceSpectrum(ConstFlux1D, amplitude=1 * units.FLAM)
    bp = spectrum.band('acs, wfc1, fr388n#3881.0')
    sp2 = constsp.normalize(10 * u.ABmag, band=bp)
    _compare_spectra(sp1, sp2)


@pytest.mark.remote_data
def test_remote_rn_calspec_u():
    sp1 = spparser.parse_spec(
        'rn(crcalspec$bd_75d325_stis_002.fits, band(u), 9.5, vegamag) * '
        'band(fos, blue, 4.3, g160l)')
    # NOTE: No expr for this combo.
    _single_functioncall(sp1, SourceSpectrum, None, '')

    bd75 = SourceSpectrum.from_file(resolve_filename(
        os.environ['PYSYN_CDBS'], 'calspec', 'bd_75d325_stis_002.fits'))
    bp_u = SpectralElement.from_filter('johnson_u')
    bd75_norm = bd75.normalize(
        9.5 * units.VEGAMAG, band=bp_u, vegaspec=spectrum.Vega)
    bp_fos = spectrum.band('fos, blue, 4.3, g160l')
    sp2 = bd75_norm * bp_fos
    _compare_spectra(sp1, sp2)


@pytest.mark.remote_data
def test_remote_z_vega():
    sp1 = spparser.parse_spec('z(crcalspec$alpha_lyr_stis_007.fits, 0.1)')
    _single_functioncall(sp1, SourceSpectrum, None,
                         'z(crcalspec$alpha_lyr_stis_007.fits,0.1)', ans_z=0.1)

    sp2 = SourceSpectrum.from_file(resolve_filename(
        os.environ['PYSYN_CDBS'], 'calspec', 'alpha_lyr_stis_007.fits'))
    sp2.z = 0.1
    _compare_spectra(sp1, sp2)


@pytest.mark.remote_data
class TestRenormPartialOverlap:
    """Test handling of ``rn(...)`` syntax for partial overlap."""
    def setup_class(self):
        self.fname = resolve_filename(
            conf.rootdir, 'etc', 'source', 'qso_fos_001.dat')

    def test_partial(self):
        """Warning only."""
        input_str = f'rn({self.fname}, band(johnson, u), 15, abmag)'
        with pytest.warns(AstropyUserWarning,
                          match=r'Spectrum is not defined everywhere'):
            sp = spparser.parse_spec(input_str)
        assert isinstance(sp, SourceSpectrum)
        assert 'force_renorm' in sp.warnings

        name = sp.meta['expr']
        assert (name.startswith('rn(') and
                name.endswith('qso_fos_001.dat,band(johnson,u),15.0,abmag)'))

    def test_disjoint(self):
        """Raise error."""
        input_str = f'rn({self.fname}, band(johnson, v), 15, abmag)'
        with pytest.raises(synexceptions.DisjointError):
            spparser.parse_spec(input_str)


@pytest.mark.remote_data
class TestEnvVar:
    """Test syntax using PYSYN_CDBS environment variable."""
    def setup_class(self):
        self.old_path = os.environ.get('PYSYN_CDBS')
        if self.old_path is None:
            os.environ['PYSYN_CDBS'] = conf.rootdir

    def test_double_slash(self):
        sp = spparser.parse_spec(
            'spec($PYSYN_CDBS//calspec/gd71_mod_005.fits)')
        assert isinstance(sp, SourceSpectrum)
        assert isinstance(sp.model, Empirical1D)

    def teardown_class(self):
        if self.old_path is None:
            del os.environ['PYSYN_CDBS']


@pytest.mark.parametrize(
    'input_str',
    ['foo(1)',
     'unit(1, nm)',
     'unit(1, vegamag)',
     'pl(5000, 1, nm)',
     'pl(5000, 1, vegamag)',
     'em(5000, 25, 1, nm)',
     'rn(bb(5000), foo(v), 17, obmag)',
     'rn(unit(1, flam), band(stis, ccd, g430m, c4451, 52X0.2), 10, abmag)',
     'rn(unit(1, flam), band(stis, ccd, mirror, 50CCD), 10, abmag)',
     'ebmvx(0.3, foo)'])
def test_parser_exception(input_str):
    """Test syntax that raises ParserError."""
    with pytest.raises(exceptions.ParserError):
        spparser.parse_spec(input_str)


class TestTokens:
    """Test underlying parser engine."""
    def setup_class(self):
        self.scanner = spparser.Scanner()

    @pytest.mark.parametrize(
        ('token_type', 'token_str'),
        [('FLOAT', '.1'),
         ('FLOAT', '1.1'),
         ('FLOAT', '1.'),
         ('FLOAT', '1'),
         ('FLOAT', '.1e+1'),
         ('FLOAT', '1.1e+1'),
         ('FLOAT', '1.e+1'),
         ('FLOAT', '1e+1'),
         ('FLOAT', '.1e-1'),
         ('FLOAT', '1.1e-1'),
         ('FLOAT', '1.e-1'),
         ('FLOAT', '1e-1'),
         ('FLOAT', '.1e1'),
         ('FLOAT', '1.1e1'),
         ('FLOAT', '1.e1'),
         ('FLOAT', '1e1'),
         ('IDENTIFIER', '/'),
         ('IDENTIFIER', 'xyzzy'),
         ('IDENTIFIER', 'xy20zzy'),
         ('IDENTIFIER', 'xyzzy20'),
         ('IDENTIFIER', '/a/b/c'),
         ('IDENTIFIER', 'foo$bar'),
         ('IDENTIFIER', 'a/b'),
         ('IDENTIFIER', '/a/b/c/foo.fits'),
         ('IDENTIFIER', 'C:/a/b/c/foo.fits')])
    def test_single_token_1(self, token_type, token_str):
        t = self.scanner.tokenize(token_str)
        assert (t[0].type, t[0].attr) == (token_type, token_str)

    @pytest.mark.parametrize(
        ('token_str', 'ans'),
        [('(', ('LPAREN', None)),
         (')', ('RPAREN', None)),
         (',', (',', None)),
         ('+', ('+', None)),
         ('*', ('*', None)),
         ('@foolist', ('FILELIST', 'foolist'))])
    def test_single_token_2(self, token_str, ans):
        t = self.scanner.tokenize(token_str)
        assert (t[0].type, t[0].attr) == ans

    @pytest.mark.parametrize(
        ('input_str', 'ans'),
        [('50CCD',
          [('FLOAT', '50'),
           ('IDENTIFIER', 'CCD')]),
         ('500X0.2',
          [('FLOAT', '500'),
           ('IDENTIFIER', 'X0.2')]),
         ('spec($PYSYN_CDBS//calspec/gd71_mod_005.fits)',
          [('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', '$PYSYN_CDBS//calspec/gd71_mod_005.fits'),
           ('RPAREN', None)]),
         ('spec(earthshine.fits) * 0.5 + '
          'rn(spec(Zodi.fits), band(johnson, v), 22.7, vegamag) + '
          '(spec(el1215a.fits) + spec(el1302a.fits) + spec(el1356a.fits) + '
          'spec(el2471a.fits)) * 0.5',
          [('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'earthshine.fits'),
           ('RPAREN', None),
           ('*', None),
           ('FLOAT', '0.5'),
           ('+', None),
           ('IDENTIFIER', 'rn'),
           ('LPAREN', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'Zodi.fits'),
           ('RPAREN', None),
           (',', None),
           ('IDENTIFIER', 'band'),
           ('LPAREN', None),
           ('IDENTIFIER', 'johnson'),
           (',', None),
           ('IDENTIFIER', 'v'),
           ('RPAREN', None),
           (',', None),
           ('FLOAT', '22.7'),
           (',', None),
           ('IDENTIFIER', 'vegamag'),
           ('RPAREN', None),
           ('+', None),
           ('LPAREN', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'el1215a.fits'),
           ('RPAREN', None),
           ('+', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'el1302a.fits'),
           ('RPAREN', None),
           ('+', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'el1356a.fits'),
           ('RPAREN', None),
           ('+', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'el2471a.fits'),
           ('RPAREN', None),
           ('RPAREN', None),
           ('*', None),
           ('FLOAT', '0.5')]),
         ('spec(earthshine.fits) * 0.5 + '
          'rn(spec(Zodi.fits), band(johnson, v), 22.7, vegamag) + '
          '(spec(el1215a.fits) * 0.1 + spec(el1302a.fits) * 0.066666667 + '
          'spec(el1356a.fits) * 0.0060 + spec(el2471a.fits) * 0.0050)',
          [('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'earthshine.fits'),
           ('RPAREN', None),
           ('*', None),
           ('FLOAT', '0.5'),
           ('+', None),
           ('IDENTIFIER', 'rn'),
           ('LPAREN', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'Zodi.fits'),
           ('RPAREN', None),
           (',', None),
           ('IDENTIFIER', 'band'),
           ('LPAREN', None),
           ('IDENTIFIER', 'johnson'),
           (',', None),
           ('IDENTIFIER', 'v'),
           ('RPAREN', None),
           (',', None),
           ('FLOAT', '22.7'),
           (',', None),
           ('IDENTIFIER', 'vegamag'),
           ('RPAREN', None),
           ('+', None),
           ('LPAREN', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'el1215a.fits'),
           ('RPAREN', None),
           ('*', None),
           ('FLOAT', '0.1'),
           ('+', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'el1302a.fits'),
           ('RPAREN', None),
           ('*', None),
           ('FLOAT', '0.066666667'),
           ('+', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'el1356a.fits'),
           ('RPAREN', None),
           ('*', None),
           ('FLOAT', '0.0060'),
           ('+', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'el2471a.fits'),
           ('RPAREN', None),
           ('*', None),
           ('FLOAT', '0.0050'),
           ('RPAREN', None)]),
         ('spec(earthshine.fits) * 0.5 + '
          'rn(spec(Zodi.fits), band(johnson, v), 22.7, vegamag) + '
          '(spec(el1215a.fits) + spec(el1302a.fits) + spec(el1356a.fits) + '
          'spec(el2471a.fits))',
          [('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'earthshine.fits'),
           ('RPAREN', None),
           ('*', None),
           ('FLOAT', '0.5'),
           ('+', None),
           ('IDENTIFIER', 'rn'),
           ('LPAREN', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'Zodi.fits'),
           ('RPAREN', None),
           (',', None),
           ('IDENTIFIER', 'band'),
           ('LPAREN', None),
           ('IDENTIFIER', 'johnson'),
           (',', None),
           ('IDENTIFIER', 'v'),
           ('RPAREN', None),
           (',', None),
           ('FLOAT', '22.7'),
           (',', None),
           ('IDENTIFIER', 'vegamag'),
           ('RPAREN', None),
           ('+', None),
           ('LPAREN', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'el1215a.fits'),
           ('RPAREN', None),
           ('+', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'el1302a.fits'),
           ('RPAREN', None),
           ('+', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'el1356a.fits'),
           ('RPAREN', None),
           ('+', None),
           ('IDENTIFIER', 'spec'),
           ('LPAREN', None),
           ('IDENTIFIER', 'el2471a.fits'),
           ('RPAREN', None),
           ('RPAREN', None)])])
    def test_composite_token(self, input_str, ans):
        t = self.scanner.tokenize(input_str)
        for expect, actual in zip(ans, t):
            assert (actual.type, actual.attr) == expect


def teardown_module():
    """Clear all cache."""
    catalog.reset_cache()
    observationmode.reset_cache()
    spectrum.reset_cache()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test observationmode.py module."""

# STDLIB
import os

# THIRD-PARTY
import numpy as np
import pytest

# ASTROPY
from astropy import units as u
from astropy.utils.data import get_pkg_data_filename

# SYNPHOT
from synphot import units

# LOCAL
from .. import observationmode
from ..config import conf
from ..stio import irafconvert

GT_FILE = get_pkg_data_filename('data/tables_tmg.fits')
CP_FILE = get_pkg_data_filename('data/tables_tmc.fits')
TH_FILE = get_pkg_data_filename('data/tables_tmt.fits')


@pytest.mark.remote_data
class TestComponent:
    """Test optical component."""
    def setup_class(self):
        self.path = irafconvert('cracscomp$')

    def test_no_interp(self):
        c = observationmode.Component(
            os.path.join(self.path, 'acs_f814w_wfc_006_syn.fits'))
        w = c.throughput.waveset[::2000]
        assert not c.empty
        assert c.throughput_name.endswith('acs_f814w_wfc_006_syn.fits')
        np.testing.assert_allclose(
            c.throughput(w).value,
            [0, 9.02462432e-07, 4.29190891e-07, 2.05288604e-01, 9.57082033e-01,
             9.38818634e-01, 4.94621272e-05, 0])

    def test_interp(self):
        c = observationmode.Component(
            os.path.join(self.path, 'acs_fr656n_006_syn.fits[fr656n#]'),
            interpval=6500)
        w = c.throughput.waveset[::50]
        assert not c.empty
        np.testing.assert_allclose(
            c.throughput(w).value,
            [9.37283281e-07, 1.01895410e-06, 1.20653385e-04, 7.53533347e-03,
             2.38073540e-03, 3.75275078e-04, 3.00504590e-06, 1.00051425e-06])

    def test_empty(self):
        c = observationmode.Component('clear')
        assert c.empty
        assert c.throughput is None


@pytest.mark.remote_data
class TestThermalComponent:
    """Test thermal component."""
    def setup_class(self):
        self.cp_name = irafconvert('crwfc3comp$wfc3_ir_win_001_syn.fits')
        self.th_name = irafconvert('crwfc3comp$wfc3_ir_win_001_th.fits')

    def test_no_interp(self):
        c = observationmode.ThermalComponent(self.cp_name, self.th_name)
        w = c.emissivity.waveset[6200:6350:10]
        assert not c.empty
        assert c.thermal_name.endswith('wfc3_ir_win_001_th.fits')
        np.testing.assert_allclose(
            c.emissivity(w).value,
            [0.03222001, 0.03887004, 0.02471, 0.02999997, 0.01115, 0.03276998,
             0.05794001], rtol=1e-6)

    def test_optical_empty(self):
        c = observationmode.ThermalComponent('clear', self.th_name)
        assert not c.empty
        assert c.throughput is None
        assert c.emissivity is not None

    def test_thermal_empty(self):
        c = observationmode.ThermalComponent(self.cp_name, 'clear')
        assert not c.empty
        assert c.throughput is not None
        assert c.emissivity is None

    def test_both_empty(self):
        c = observationmode.ThermalComponent('clear', 'clear')
        assert c.empty
        assert c.throughput is None
        assert c.emissivity is None


class TestProcessGraphTable:
    """Test graph table processing logic."""
    @pytest.mark.remote_data
    def test_default(self):
        gt, gtname, area = observationmode._process_graphtable(None)
        assert gtname.endswith('tmg.fits')  # Latest from CDBS
        assert area == conf.area * units.AREA

    def test_other_logic(self):
        gt_file = get_pkg_data_filename('data/tables_primarea_tmg.fits')
        gt, gtname, area = observationmode._process_graphtable(gt_file)
        assert area == 100 * units.AREA

        # Cached data
        gt2, gtname2, area2 = observationmode._process_graphtable(gt_file)
        assert gt2 is gt
        assert gtname2 == gtname
        assert area2 == area


@pytest.mark.remote_data
class TestObservationMode:
    """Test observation mode."""
    def setup_class(self):
        observationmode.reset_cache()
        self.obsmode = observationmode.ObservationMode(
            'acs,wfc1,f555w,mjd#56597',
            graphtable=GT_FILE, comptable=CP_FILE)

    def test_cache(self):
        gt_key = list(observationmode._GRAPHDICT.keys())[0]
        assert gt_key.endswith('tables_tmg.fits')

        cp_key = list(observationmode._COMPDICT.keys())[0]
        assert cp_key.endswith('tables_tmc.fits')

        det_key = list(observationmode._DETECTORDICT.keys())[0]
        assert det_key.endswith('detectors.dat')

    def test_base_attributes(self):
        """Some graph, component, and wave table behaviors are tested
        somewhere else, so not repeated here.

        """
        assert self.obsmode.modes == ['acs', 'wfc1', 'f555w', 'mjd#']
        assert self.obsmode.pardict == {'mjd': 56597}
        assert self.obsmode.ctname.endswith('tables_tmc.fits')
        assert self.obsmode.pixscale == 0.05 * u.arcsec
        assert str(self.obsmode) == 'acs,wfc1,f555w,mjd#56597'
        assert len(self.obsmode) == 5
        np.testing.assert_allclose(
            self.obsmode._constant.value, 2277381155648.759, rtol=1e-6)

    def test_components(self):
        for c in self.obsmode.components:
            assert not c.empty

    def test_throughput(self):
        t = self.obsmode.throughput
        w = t.waveset[::1000]
        np.testing.assert_allclose(
            w.value,
            [500, 3797, 4797, 5797, 6797, 7797, 8797, 9797, 10796])
        np.testing.assert_allclose(
            t(w).value,
            [0, 6.87477055e-06, 2.14807644e-01, 2.85725420e-01, 7.34488102e-07,
             2.96078061e-08, 3.57507695e-08, 9.99538418e-07, 0], rtol=1e-3)

    def test_sensitivity(self):
        sens = self.obsmode.sensitivity
        w = sens.waveset[::1000]
        np.testing.assert_allclose(
            w.value,
            [500, 3797, 4797, 5797, 6797, 7797, 8797, 9797, 10796])
        np.testing.assert_allclose(
            sens(w).value,
            [0, 5.94476276e+10, 2.34668703e+15, 3.77214087e+15, 1.13694055e+10,
             5.25738140e+08, 7.16235278e+08, 2.23012046e+10, 0], rtol=1e-3)

    def test_interp_two_params(self):
        obsmode = observationmode.ObservationMode(
            'acs,hrc,fr459m#4610,aper#0.3',
            graphtable=GT_FILE, comptable=CP_FILE)
        assert obsmode.modes == ['acs', 'hrc', 'fr459m#', 'aper#']
        assert obsmode.pardict['aper'] == 0.3
        assert obsmode.pardict['fr459m'] == 4610

    def test_thermal_spec(self):
        """Also see TestThermalObservationMode.
        Whitespace in obsmode should not matter.
        Flux values are from ASTROLIB PYSYNPHOT, except the first value,
        which is supposed to be zero.

        """
        obsmode = observationmode.ObservationMode(
            'wfc3, ir, f153m', graphtable=GT_FILE, comptable=CP_FILE)
        thsp = obsmode.thermal_spectrum(thermtable=TH_FILE)
        w = [6898, 7192, 7486, 7780, 8630, 11190, 13790, 15670, 17954.90234375]
        np.testing.assert_allclose(
            thsp(w).value,
            [2.911303e-30, 9.52169734e-29, 2.32066138e-27, 4.38458258e-26,
             1.86992105e-21, 2.90270169e-15, 3.42776351e-11, 2.15179718e-08,
             7.24300701e-09], rtol=5e-3)


@pytest.mark.remote_data
class TestThermalObservationMode:
    """Test thermal observation mode.

    .. note:: to_spectrum() method is tested in TestObservationMode.

    """
    def setup_class(self):
        self.thmode = observationmode.ThermalObservationMode(
            'wfc3,ir,f153m', graphtable=GT_FILE, comptable=CP_FILE,
            thermtable=TH_FILE)

    def test_cache(self):
        th_key = list(observationmode._THERMDICT.keys())[0]
        assert th_key.endswith('tables_tmt.fits')

    def test_attributes(self):
        assert str(self.thmode) == 'wfc3,ir,f153m (thermal)'

    def test_components(self):
        ans = ['wfc3_ir_primary_001_th.fits',
               'wfc3_ir_pads_001_th.fits',
               'wfc3_ir_secondary_001_th.fits',
               'wfc3_pom_001_th.fits',
               'wfc3_ir_csm_001_th.fits',
               'wfc3_ir_fold_001_th.fits',
               'wfc3_ir_mir1_001_th.fits',
               'wfc3_ir_mir2_001_th.fits',
               'wfc3_ir_mask_001_th.fits',
               'wfc3_ir_rcp_001_th.fits',
               'wfc3_ir_rcp_001_th.fits',
               'wfc3_ir_f153m_002_th.fits',
               'wfc3_ir_wmring_001_th.fits',
               'wfc3_ir_win_001_th.fits',
               'wfc3_ir_qe_003_th.fits',
               'clear']
        for c, a in zip(self.thmode.components, ans):
            s = c.thermal_name
            assert s.endswith(a)
            if s == 'clear':
                assert c.emissivity is None
            else:
                assert c.emissivity is not None

    def test_exceptions(self):
        with pytest.raises(NotImplementedError):
            observationmode.ThermalObservationMode(
                'acs,wfc1,f555w', graphtable=GT_FILE, comptable=CP_FILE,
                thermtable=TH_FILE)


def teardown_module():
    observationmode.reset_cache()
    assert observationmode._GRAPHDICT == {}



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test wavetable.py module."""

# THIRD-PARTY
import numpy as np
import pytest

# ASTROPY
from astropy import units as u

# LOCAL
from .. import exceptions
from ..wavetable import WAVECAT


@pytest.mark.parametrize(
    ('obsmode', 'ans'),
    [('nicmos,3,f220m', '(7000.0,29996.0,1.0)'),
     ('acs,hrc,f550m', WAVECAT['acs,hrc']),
     ('stis,ccd,g750m', WAVECAT['stis,g750m']),
     ('stis,fuvmama,g140l,s52x2', WAVECAT['stis,g140l']),
     ('stis,nuvmama,e230h,c2263,s02x02', WAVECAT['stis,e230h,c2263'])])
def test_getitem(obsmode, ans):
    """Waveset catalog access."""
    assert WAVECAT[obsmode] == ans


def test_getitem_exceptions():
    """Waveset catalog exceptions."""
    with pytest.raises(KeyError):
        WAVECAT['johnson,v']
    with pytest.raises(exceptions.AmbiguousObsmode):
        WAVECAT['acs,wfc1,wfc2']


def test_load_waveset_file():
    """Load waveset from file."""
    par, wave = WAVECAT.load_waveset('acs,wfc1')
    assert par.endswith('wavecats/acs.dat')
    assert wave.unit == u.AA
    np.testing.assert_allclose([wave.value[0], wave.value[-1]], [1000, 11000])


@pytest.mark.parametrize(
    ('obsmode', 'ncoeff', 'ans'),
    [('stis,g230l', 3, [1568, 3184]),
     ('wfc3,uvis2,g280', 4, [1850, 9998])])
def test_load_waveset_coeff(obsmode, ncoeff, ans):
    """Load waveset from coefficients."""
    par, wave = WAVECAT.load_waveset(obsmode)
    assert len(par[1:-1].split(',')) == ncoeff
    assert wave.unit == u.AA
    np.testing.assert_allclose([wave.value[0], wave.value[-1]], ans)



"""This module contains ACS/SBC commissioning tests.
Adapted from ``astrolib/pysynphot/from_commissioning/acs/test3.py``
(last few tests).

.. note::

    ``astrolib/pysynphot/from_commissioning/acs/test4.py`` was disabled.
    Therefore, Test622 to Test 657 were not ported over.

"""

# LOCAL
from ..utils import CommCase


class Test619(CommCase):
    obsmode = 'acs,sbc,f115lp'
    spectrum = 'rn(unit(1.0,flam),band(johnson,v),15,vegamag)'


class Test620(Test619):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.5e-16,flam)'


class Test621(Test619):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1e-18,flam)'



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Custom exceptions for `stsynphot` to raise."""

# SYNPHOT
from synphot.exceptions import SynphotError

__all__ = ['ParserError', 'GenericASTTraversalPruningException',
           'ParameterOutOfBounds', 'GraphtabError', 'UnusedKeyword',
           'IncompleteObsmode', 'AmbiguousObsmode']


class ParserError(SynphotError):
    """Exceptions for language parser."""
    pass


class GenericASTTraversalPruningException(ParserError):
    """SPARK AST traversal pruning exception."""
    pass


class ParameterOutOfBounds(SynphotError):
    """Exceptions for catalog problems."""
    pass


class GraphtabError(SynphotError):
    """Exceptions to do with graph table traversal."""
    pass


class UnusedKeyword(GraphtabError):
    """Unused keyword is not allowed in graph table."""
    pass


class IncompleteObsmode(GraphtabError):
    """Incomplete observation mode is not allowed in graph table."""
    pass


class AmbiguousObsmode(GraphtabError):
    """Ambiguous observation mode is not allowed in graph table."""
    pass


class PixscaleNotFoundError(SynphotError):
    """Undefined pixel scale for a given observation mode."""
    pass



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test config.py module."""

# STDLIB
import os

# THIRD-PARTY
import numpy as np
import pytest

# SYNPHOT
from synphot.config import conf as synconf
from synphot.utils import generate_wavelengths

# LOCAL
from .. import config
from ..stio import get_latest_file, irafconvert


class TestOverwriteSynphot:
    """Test if overwriting ``synphot`` defaults is successful."""
    def setup_class(self):
        # For some reason, this does not automatically execute during testing.
        config.overwrite_synphot_config(config.conf.rootdir)

        self.vegafile = synconf.vega_file

    def test_dirname(self):
        assert self.vegafile.startswith(config.conf.rootdir)

    @pytest.mark.remote_data
    def test_isfile(self):
        if self.vegafile.startswith(('ftp', 'http')):
            # This is the case on Travis CI
            pytest.xfail('Cannot test this over FTP or HTTP')
        else:
            assert os.path.isfile(self.vegafile)


@pytest.mark.remote_data
class TestConfigChanges:
    def setup_class(self):
        self.def_dict = config.getref()

    @pytest.mark.parametrize(
        ('cfgname', 'new_val'),
        [('graphtable', 'mtab$n9i1408hm_tmg.fits'),
         ('comptable', 'mtab$n9i1408im_tmc.fits'),
         ('thermtable', 'mtab$n5k15531m_tmt.fits')])
    def test_tables_area(self, cfgname, new_val):
        # Same as config.conf.cfgname = new_val
        setattr(config.conf, cfgname, new_val)
        assert getattr(config.conf, cfgname) == new_val

        # Reset to default
        config.conf.reset(cfgname)
        old_expanded_val = get_latest_file(irafconvert(
            getattr(config.conf, cfgname)))
        assert old_expanded_val == self.def_dict[cfgname]

    def test_area(self):
        config.conf.area = 1
        assert config.conf.area == 1

        # Reset to default
        config.conf.reset('area')
        assert config.conf.area == self.def_dict['area']

    def test_waveset(self):
        w = generate_wavelengths(
            minwave=3000, maxwave=5000, num=100, log=False)
        config.conf.waveset_array = w[0].value.tolist()
        config.conf.waveset = w[1]
        np.testing.assert_allclose(
            [config.conf.waveset_array[0], config.conf.waveset_array[-1]],
            [3000, 4980])
        assert (config.conf.waveset ==
                'Min: 3000, Max: 5000, Num: 100, Delta: None, Log: False')

        # Reset to default
        config.conf.reset('waveset_array')
        config.conf.reset('waveset')
        np.testing.assert_allclose(
            [config.conf.waveset_array[0], config.conf.waveset_array[-1]],
            [500, 25989.72879567])
        assert config.conf.waveset == self.def_dict['waveset']



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This module contains spectrum classes specific to STScI formats."""

# STDLIB
import os
import re
import warnings

# THIRD-PARTY
import numpy as np

# ASTROPY
from astropy import log
from astropy import units as u
from astropy.utils.exceptions import AstropyUserWarning

# SYNPHOT
from synphot import binning, reddening, units
from synphot import exceptions as synexceptions
from synphot.config import conf as synconf
from synphot.models import Empirical1D
from synphot.spectrum import SourceSpectrum, SpectralElement

# LOCAL
from . import stio
from .exceptions import PixscaleNotFoundError

__all__ = ['reset_cache', 'interpolate_spectral_element',
           'ObservationSpectralElement', 'band', 'ebmvx', 'load_vega', 'Vega']

_interpfilepatt = re.compile(r'\[(?P<col>.*?)\]')
_REDLAWS = {}  # Cache previously loaded reddening laws
Vega = None  # Cache Vega spectrum


def reset_cache():
    """Empty the reddening laws cache."""
    global _REDLAWS
    _REDLAWS.clear()


def _interp_spec(interpval, waves, lower_val, upper_val, lower_thru,
                 upper_thru, doshift):
    """Interpolate between two spectra to parameter at given value.
    Called by :func:`interpolate_spectral_element`.

    Parameters
    ----------
    interpval : float
        Desired parameter value.

    waves : array-like
        Wavelengths for all the parameterized spectra.

    lower_val, upper_val : float
        Low and high values of parameter to interpolate.

    lower_thru, upper_thru : array-like
        Throughput values are ``lower_val`` and ``upper_val``.

    doshift : bool
        Perform wavelength shift before interpolation.

    Returns
    -------
    thru : array-like
        Interpolated throughput at ``interpval``.

    """
    if doshift:
        # Adjust the wavelength table to bracket the range
        lwave = waves + (lower_val - interpval)
        uwave = waves + (upper_val - interpval)

        # Interpolate the columns at those ranges
        lower_thru = np.interp(lwave, waves, lower_thru)
        upper_thru = np.interp(uwave, waves, upper_thru)

    # Then interpolate between the two columns
    w = (interpval - lower_val) / (upper_val - lower_val)
    return (upper_thru * w) + lower_thru * (1.0 - w)


def _extrap_spec(interpval, lower_val, upper_val, lower_thru, upper_thru):
    """Extrapolate using two spectra to parameter at given value.
    Called by :func:`interpolate_spectral_element`.
    Also see :func:`_interp_spec`.

    """
    m = (upper_thru - lower_thru) / (upper_val - lower_val)
    b = lower_thru - m * lower_val
    return m * interpval + b


def interpolate_spectral_element(parfilename, interpval, ext=1):
    """Interpolate (or extrapolate) throughput spectra in given
    parameterized FITS table to given parameter value.

    FITS table is parsed with :func:`stsynphot.stio.read_interp_spec`.
    Parameterized values must be in ascending order in the
    table columns.

    If extrapolation is needed but not allowed, default throughput
    from ``THROUGHPUT`` column will be used.

    Parameters
    ----------
    parfilename : str
        Parameterized filename contains a suffix followed by
        a column name specificationin between square brackets.
        For example, ``path/acs_fr656n_006_syn.fits[fr656n#]``.

    interpval : float
        Desired parameter value.

    ext : int, optional
        FITS extension index of the data table.

    Returns
    -------
    sp : `synphot.spectrum.SpectralElement`
        Empirical bandpass at ``interpval``.

    Raises
    ------
    synphot.exceptions.ExtrapolationNotAllowed
        Extrapolation is not allowed by data table.

    synphot.exceptions.SynphotError
        No columns available for interpolation or extrapolation.

    """
    def_colname = 'THROUGHPUT'
    warndict = {}

    # Separate real filename and column name specification
    xre = _interpfilepatt.search(parfilename)
    if xre is None:
        raise synexceptions.SynphotError(
            f'{parfilename} must be in the format of "path/filename.fits'
            '[col#]"')
    filename = parfilename[0:xre.start()]
    col_prefix = xre.group('col').upper()

    # Read data table
    try:
        data, wave_unit, doshift, extrapolate = stio.read_interp_spec(
            filename, tab_ext=ext)
    except Exception as e:  # pragma: no cover
        raise IOError(f'Failed to read {filename}[{ext}]: {repr(e)}')
    wave_unit = units.validate_unit(wave_unit)
    wave0 = data['WAVELENGTH']

    # Determine the columns that bracket the desired value.
    # Grab all columns that begin with the parameter name (e.g. 'MJD#')
    # and then split off the numbers after the '#'.
    col_names = []
    col_pars = []
    for n in data.names:
        cn = n.upper()
        if cn.startswith(col_prefix):
            col_names.append(cn)
            col_pars.append(float(cn.split('#')[1]))

    if len(col_names) < 1:
        raise synexceptions.SynphotError(
            f'{filename} contains no interpolated columns for {col_prefix}.')

    # Assumes ascending order of parameter values in table.
    min_par = col_pars[0]
    max_par = col_pars[-1]

    # Exact match. No interpolation needed.
    if interpval in col_pars:
        thru = data[col_names[col_pars.index(interpval)]]

    # Need interpolation.
    elif (interpval > min_par) and (interpval < max_par):
        upper_ind = np.searchsorted(col_pars, interpval)
        lower_ind = upper_ind - 1

        thru = _interp_spec(
            interpval, wave0, col_pars[lower_ind], col_pars[upper_ind],
            data[col_names[lower_ind]], data[col_names[upper_ind]], doshift)

    # Need extrapolation, if allowed.
    elif extrapolate:
        # Extrapolate below lowest columns.
        if interpval < min_par:
            thru = _extrap_spec(interpval, min_par, col_pars[1],
                                data[col_names[0]], data[col_names[1]])

        # Extrapolate above highest columns.
        else:  # interpval > max_par
            thru = _extrap_spec(interpval, col_pars[-2], max_par,
                                data[col_names[-2]], data[col_names[-1]])

    # Extrapolation not allowed.
    else:
        # Use default, if available.
        if def_colname in data.names:
            warnings.warn(
                'Extrapolation not allowed, using default throughput for '
                f'{parfilename}.', AstropyUserWarning)
            warndict['DefaultThroughput'] = True
            thru = data[def_colname]

        # Nothing can be done.
        else:
            raise synexceptions.ExtrapolationNotAllowed(
                f'No default throughput for {parfilename}.')

    meta = {'expr': f'{filename}#{interpval:g}',
            'warnings': warndict}
    return SpectralElement(
        Empirical1D, points=wave0 * wave_unit, lookup_table=thru, meta=meta)


class ObservationSpectralElement(SpectralElement):
    """Class to handle bandpass from observation mode.

    This class has additional methods that are specific
    to observation mode and instrument-specific wavelength set.

    .. note::

        For methods that take ``area``, it is *recommended* but
        not required to use `area`.

    Parameters
    ----------
    modelclass, kwargs
        See `~synphot.spectrum.BaseSpectrum`.

    obsmode : `~stsynphot.observationmode.ObservationMode`
        Observation mode for this bandpass.

    """
    def __init__(self, modelclass, obsmode=None, **kwargs):
        if obsmode is None:
            raise synexceptions.SynphotError('Missing OBSMODE.')

        super(ObservationSpectralElement, self).__init__(modelclass, **kwargs)
        self._obsmode = obsmode
        self.meta['expr'] = str(obsmode)

        # Check for zero bounds, if applicable
        try:
            self.bounded_by_zero()
        except synexceptions.SynphotError:  # pragma: no cover
            warnings.warn(
                'Zero-bound check not done due to undefined waveset.',
                AstropyUserWarning)

    @property
    def obsmode(self):
        """Observation mode for this bandpass."""
        return self._obsmode

    @property
    def binset(self):
        """Instrument-specific wavelength set from
        ``stsynphot.wavetable.WAVECAT`` based on `obsmode`."""
        return self.obsmode.bandwave

    @property
    def area(self):
        """Telescope collecting area based on `obsmode`."""
        return self.obsmode.primary_area

    def __len__(self):
        """Get the number of components in `obsmode`."""
        return len(self.obsmode)

    def taper(self, **kwargs):
        """Disabled."""
        raise NotImplementedError('Method disabled.')

    def bounded_by_zero(self, wavelengths=None, verbose=True):
        """Check if sampled throughout is bounded by zeroes.

        Parameters
        ----------
        wavelengths : array-like, `~astropy.units.quantity.Quantity`, or `None`
            Wavelength values for sampling.
            If not a Quantity, assumed to be in Angstrom.
            If `None`, ``self.waveset`` is used.

        verbose : bool
            Print warning if unbounded.

        Returns
        -------
        bounded : bool
            `True` if bounded by zeroes, `False` otherwise.

        """
        thru = self(wavelengths)
        y = thru[::thru.size - 1].value
        bounded = np.all(y == 0)

        if not bounded and verbose:
            warnings.warn(
                f'Unbounded throughput; {y} when expecting zeroes.',
                AstropyUserWarning)

        return bounded

    def showfiles(self):  # pragma: no cover
        """Display ``self.obsmode`` optical component filenames.

        .. note:: Similar to IRAF SYNPHOT SHOWFILES.

        """
        self.obsmode.showfiles()

    def thermback(self, area=None, thermtable=None):
        """Calculate thermal background count rate for
        ``self.obsmode``.

        Calculation uses
        :func:`~stsynphot.observationmode.ObservationMode.thermal_spectrum`
        to extract thermal component source spectrum in
        PHOTLAM per square arcsec. Then this spectrum is
        integrated and multiplied by detector pixel scale
        and telescope collecting area to produce a count rate
        in count/s/pix. This unit is non-standard but used widely
        by STScI Exposure Time Calculator.

        .. note::

            Similar to IRAF SYNPHOT THERMBACK.

        Parameters
        ----------
        area : float, `~astropy.units.quantity.Quantity`, or `None`
            Area that flux covers.
            If not a Quantity, assumed to be in :math:`cm^{2}`.
            If `None`, use `area`.

        thermtable : str or `None`
            Thermal component table filename.
            If `None`, uses ``stsynphot.config.conf.thermtable``.

        Returns
        -------
        bg : `~astropy.units.quantity.Quantity`
            Thermal background count rate.

        Raises
        ------
        stsynphot.exceptions.PixscaleNotFoundError
            Undefined pixel scale for the given observation mode.

        """
        if self.obsmode.pixscale is None:
            raise PixscaleNotFoundError(
                f'Undefined pixel scale for {self.obsmode}.')

        if area is None:
            area = self.area
        area = units.validate_quantity(area, units.AREA)

        sp = self.obsmode.thermal_spectrum(thermtable=thermtable)
        bg = sp.integrate() * self.obsmode.pixscale ** 2 * area

        return bg.value * (u.count / u.s / u.pix)

    def binned_waverange(self, cenwave, npix, **kwargs):
        """Calculate the wavelength range covered by the given number
        of pixels centered on the given central wavelengths of `binset`.

        Parameters
        ----------
        cenwave : float or `~astropy.units.quantity.Quantity`
            Desired central wavelength.
            If not a Quantity, assumed to be in Angstrom.

        npix : int
            Desired number of pixels, centered on ``cenwave``.

        kwargs : dict
            Keywords accepted by :func:`synphot.binning.wave_range`.

        Returns
        -------
        waverange : `~astropy.units.quantity.Quantity`
            Lower and upper limits of the wavelength range,
            in the unit of ``cenwave``.

        Raises
        ------
        synphot.exceptions.UndefinedBinset
            Undefined `binset`.

        """
        if self.binset is None:
            raise synexceptions.UndefinedBinset(
                'No binset specified for this passband.')

        # Calculation is done in the unit of cenwave.
        if not isinstance(cenwave, u.Quantity):
            cenwave = cenwave * self._internal_wave_unit

        bin_wave = units.validate_quantity(
            self.binset, cenwave.unit, equivalencies=u.spectral())

        return binning.wave_range(
            bin_wave.value, cenwave.value, npix, **kwargs) * cenwave.unit

    def binned_pixelrange(self, waverange, **kwargs):
        """Calculate the number of pixels within the given wavelength
        range and `binset`.

        Parameters
        ----------
        waverange : tuple of float or `~astropy.units.quantity.Quantity`
            Lower and upper limits of the desired wavelength range.
            If not a Quantity, assumed to be in Angstrom.

        kwargs : dict
            Keywords accepted by :func:`synphot.binning.pixel_range`.

        Returns
        -------
        npix : number
            Number of pixels.

        Raises
        ------
        synphot.exceptions.UndefinedBinset
            Undefined `binset`.

        """
        if self.binset is None:
            raise synexceptions.UndefinedBinset(
                'No binset specified for this passband.')

        x = units.validate_quantity(
            waverange, self._internal_wave_unit, equivalencies=u.spectral())

        return binning.pixel_range(self.binset.value, x.value, **kwargs)

    def to_fits(self, filename, wavelengths=None, **kwargs):
        """Write the spectrum to a FITS file.

        Throughput column is automatically named 'THROUGHPUT'.
        Graph and optical component tables are written to
        table header (not primary) under these keywords:

            * ``GRFTABLE``
            * ``CMPTABLE``

        Parameters
        ----------
        filename : str
            Output filename.

        wavelengths : array-like, `~astropy.units.quantity.Quantity`, or `None`
            Wavelength values for sampling.
            If not a Quantity, assumed to be in Angstrom.
            If `None`, ``self.waveset`` is used.

        kwargs : dict
            Keywords accepted by :func:`synphot.specio.write_fits_spec`.

        """
        bkeys = {
            'grftable': (os.path.basename(self.obsmode.gtname),
                         'graph table used'),
            'cmptable': (os.path.basename(self.obsmode.ctname),
                         'component table used')}

        if 'ext_header' in kwargs:
            kwargs['ext_header'].update(bkeys)
        else:
            kwargs['ext_header'] = bkeys

        super(ObservationSpectralElement, self).to_fits(
            filename, wavelengths=wavelengths, **kwargs)

    @classmethod
    def from_file(cls, filename, **kwargs):
        """Disabled."""
        raise NotImplementedError('Class method disabled.')

    @classmethod
    def from_filter(cls, filtername, **kwargs):
        """Disabled."""
        raise NotImplementedError('Class method disabled.')

    @classmethod
    def from_obsmode(cls, obsmode, graphtable=None, comptable=None,
                     component_dict={}):
        """Create a bandpass from observation mode string.

        Parameters
        ----------
        obsmode : str
            Observation mode.

        graphtable : str or `None`
            Graph table filename.
            If `None`, uses ``stsynphot.config.conf.graphtable``.

        comptable : str or `None`
            Optical component table filename.
            If `None`, uses ``stsynphot.config.conf.comptable``.

        component_dict : dict
            Maps component filename to corresponding
            `~stsynphot.observationmode.Component`.

        Returns
        -------
        bp : `ObservationSpectralElement`
            Empirical bandpass.

        Raises
        ------
        synphot.exceptions.SynphotError
            Observation mode yields no throughput.

        """
        from .observationmode import ObservationMode  # Avoid circular import

        ob = ObservationMode(
            obsmode, graphtable=graphtable, comptable=comptable,
            component_dict=component_dict)

        if not isinstance(ob.throughput, SpectralElement):  # pragma: no cover
            raise synexceptions.SynphotError(f'{obsmode} has no throughput.')

        return cls(ob.throughput, obsmode=ob)


def band(*args, **kwargs):
    """Convenience function to create a bandpass with
    an observation mode string.

    See :func:`ObservationSpectralElement.from_obsmode` and
    :ref:`stsynphot-obsmode`.

    """
    return ObservationSpectralElement.from_obsmode(*args, **kwargs)


def ebmvx(redlaw_name, ebv):
    """Convenience function to create extinction curve for
    given reddening law and :math:`E(B-V)`.

    Parameters
    ----------
    redlaw_name : str
        Reddening law model name
        (see :meth:`synphot.reddening.ReddeningLaw.from_extinction_model`).
        Choose from 'lmc30dor', 'lmcavg', 'mwavg', 'mwdense', 'mwrv21',
        'mwrv40', 'smcbar', 'xgalsb', 'gal3', or `None`.
        ``gal3`` and `None` are same as ``mwavg`` and used for testing only.

    ebv : float or `~astropy.units.quantity.Quantity`
        :math:`E(B-V)` value in magnitude. See
        :meth:`synphot.reddening.ReddeningLaw.extinction_curve`.

    Returns
    -------
    extcurve : `synphot.reddening.ExtinctionCurve`
        Extinction curve.

    """
    global _REDLAWS

    if redlaw_name in ('gal3', None):
        m = 'mwavg'
        log.info(f'{redlaw_name} uses {m} reddening law.')
    else:
        m = redlaw_name

    if m not in _REDLAWS:
        _REDLAWS[m] = reddening.ReddeningLaw.from_extinction_model(
            m, encoding='binary')

    return _REDLAWS[m].extinction_curve(ebv)


def load_vega(vegafile=None, **kwargs):
    """Convenience function to load Vega spectrum that is
    used throughout `stsynphot`.

    Parameters
    ----------
    vegafile : str or `None`, optional
        Vega spectrum filename.
        If `None`, use ``synphot.config.conf.vega_file``.

    kwargs : dict
        Keywords acceptable by :func:`synphot.specio.read_remote_spec`.

    Returns
    -------
    sp : `synphot.spectrum.SourceSpectrum` or `None`
        Vega spectrum. `None` if failed.

    """
    global Vega

    if vegafile is None:
        vegafile = synconf.vega_file

    with synconf.set_temp('vega_file', vegafile):
        try:
            Vega = SourceSpectrum.from_vega(**kwargs)
        except Exception as e:
            Vega = None
            warnings.warn(
                f'Failed to load Vega spectrum from {vegafile}; Functionality '
                f'involving Vega will be cripped: {repr(e)}',
                AstropyUserWarning)


# Load default Vega
load_vega(encoding='binary')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Synthetic photometry language parser.

See :ref:`stsynphot-parser` for more details and
:func:`BaseParser.p_top` for language definition.

.. note::

    Like `stsynphot.spark`, parser docstrings in this module
    are used by the parser itself, so modify any docstring in
    this module with care.

    In `Interpreter`, the docstring of every function named with ``p_*``
    is part of the instructions to the parser.

    IRAF SYNPHOT extinction names are obsolete and no longer supported.

"""
# ASTROPY
from astropy import log
from astropy import units as u

# SYNPHOT
from synphot import exceptions as synexceptions
from synphot import units
from synphot.models import (BlackBodyNorm1D, Box1D, ConstFlux1D,
                            GaussianFlux1D, PowerLawFlux1D)
from synphot.spectrum import SourceSpectrum, SpectralElement

# LOCAL
from . import exceptions, spectrum
from .catalog import grid_to_spec
from .config import conf
from .spark import GenericScanner, GenericASTBuilder, GenericASTMatcher
from .stio import irafconvert

__all__ = ['Token', 'AST', 'BaseScanner', 'Scanner', 'BaseParser',
           'Interpreter', 'tokens_info', 'scan', 'parse', 'interpret',
           'parse_spec']

# IRAF SYNPHOT functions
_SYFUNCTIONS = ('band', 'bb', 'box', 'ebmvx', 'em', 'icat', 'pl', 'rn', 'spec',
                'unit', 'z')

# IRAF SYNPHOT flux units
_SYFORMS = ('abmag', 'counts', 'flam', 'fnu', 'jy', 'mjy', 'obmag', 'photlam',
            'photnu', 'stmag', 'vegamag')


def _convertstr(value):
    """Convert given filename to source spectrum or passband.

    This is used by the interpreter to do the conversion from
    string to spectrum object.

    """
    if not isinstance(value, str):
        return value
    value = irafconvert(value)
    try:
        sp = SourceSpectrum.from_file(value)
    except KeyError:
        sp = SpectralElement.from_file(value)
    return sp


class Token:
    # Class to handle token.
    def __init__(self, token_type=None, attr=None):
        self.type = token_type
        self.attr = attr

    def __cmp__(self, o):  # pragma: py2
        return cmp(self.type, o)

    def __eq__(self, o):  # pragma: py3
        return self.type == o

    def __lt__(self, o):  # pragma: py3
        return self.type < o

    def __repr__(self):
        if self.attr is not None:
            return str(self.attr)
        else:
            return self.type


class AST:
    # Class to handle Abstract Syntax Tree (AST).
    def __init__(self, ast_type):
        self.type = ast_type
        self._kids = []

    def __getitem__(self, i):
        return self._kids[i]

    def __setitem__(self, i, seq):  # pragma: py3
        self._kids[i] = seq

    def __len__(self):
        return len(self._kids)

    def __setslice__(self, low, high, seq):  # pragma: py2
        self._kids[low:high] = seq

    def __cmp__(self, o):  # pragma: py2
        return cmp(self.type, o)

    def __eq__(self, o):  # pragma: py3
        return self.type == o

    def __lt__(self, o):  # pragma: py3
        return self.type < o


class BaseScanner(GenericScanner):
    # Base class to handle language scanner.
    def tokenize(self, s):
        # Tokenize string.
        self.rv = []
        GenericScanner.tokenize(self, s)
        return self.rv

    def t_whitespace(self, s):
        # Whitespace regular expression.
        r' \s+ '

    def t_op(self, s):
        # Addition, multiplication, and subtraction operations
        # regular expression.
        r' \+ | \* | - '
        self.rv.append(Token(token_type=s))

    def t_lparens(self, s):
        # Left parenthesis regular expression.
        r' \( '
        self.rv.append(Token(token_type='LPAREN'))  # nosec

    def t_rparens(self, s):
        # Right parenthesis regular expression.
        r' \) '
        self.rv.append(Token(token_type='RPAREN'))  # nosec

    def t_comma(self, s):
        # Comma regular expression.
        r' , '
        self.rv.append(Token(token_type=s))

    def t_integer(self, s):
        # Integer regular expression.
        r' \d+ '
        self.rv.append(Token(token_type='INTEGER', attr=s))  # nosec

    def t_identifier(self, s):
        # Identifier regular expression.
        r' [$a-z_A-Z/\//][\w/\.\$:#]*'
        self.rv.append(Token(token_type='IDENTIFIER', attr=s))  # nosec

    def t_filelist(self, s):
        # File list regular expression.
        r' @\w+'
        self.rv.append(Token(token_type='FILELIST', attr=s[1:]))  # nosec


class Scanner(BaseScanner):
    # Class to handle language scanner.
    def t_float(self, s):
        # Float regular expression.
        r' ((\d*\.\d+)|(\d+\.d*)|(\d+)) ([eE][-+]?\d+)?'
        self.rv.append(Token(token_type='FLOAT', attr=s))  # nosec

    def t_divop(self, s):
        # Division operation regular expression.
        r' \s/\s '
        self.rv.append(Token(token_type='/'))  # nosec


class BaseParser(GenericASTBuilder):
    # Base class to handle language parser.
    def __init__(self, ASTclass, start='top'):
        super(BaseParser, self).__init__(ASTclass, start)

    def p_top(self, args):
        """
            top ::= expr
            top ::= FILELIST
            expr ::= expr + term
            expr ::= expr - term
            expr ::= term
            term ::= term * factor
            term ::= term / factor
            value ::= LPAREN expr RPAREN
            term ::= factor
            factor ::= unaryop value
            factor ::= value
            unaryop ::= +
            unaryop ::= -
            value ::= INTEGER
            value ::= FLOAT
            value ::= IDENTIFIER
            value ::= function_call
            function_call ::= IDENTIFIER LPAREN arglist RPAREN
            arglist ::= arglist , expr
            arglist ::= expr
        """

    def terminal(self, token):
        # Return terminal element.
        rv = AST(token.type)
        rv.attr = token.attr
        return rv

    def nonterminal(self, intype, args):
        # Return non-terminal element.
        if len(args) == 1:
            rv = args[0]
        else:
            rv = GenericASTBuilder.nonterminal(self, intype, args)
        return rv


class Interpreter(GenericASTMatcher):
    # Class to handle language interpreter.
    def __init__(self, ast):
        super(Interpreter, self).__init__('V', ast)

    def error(self, token):
        # Raise en exception.
        raise exceptions.ParserError('Cannot interpret AST.')

    def p_int(self, tree):
        """ V ::= INTEGER """
        tree.value = int(tree.attr)
        tree.svalue = tree.attr

    def p_float(self, tree):
        """ V ::= FLOAT """
        tree.value = float(tree.attr)
        tree.svalue = tree.attr

    def p_identifier(self, tree):
        """ V ::= IDENTIFIER """
        tree.value = tree.attr
        tree.svalue = tree.attr

    def p_factor_unary_plus(self, tree):
        """ V ::= factor ( + V ) """
        tree.value = _convertstr(tree[1].value)

    def p_factor_unary_minus(self, tree):
        """ V ::= factor ( - V ) """
        tree.value = - _convertstr(tree[1].value)

    def p_expr_plus(self, tree):
        """ V ::= expr ( V + V ) """
        tree.value = _convertstr(tree[0].value) + _convertstr(tree[2].value)

    def p_expr_minus(self, tree):
        """ V ::= expr ( V - V ) """
        tree.value = _convertstr(tree[0].value) - _convertstr(tree[2].value)

    def p_term_mult(self, tree):
        """ V ::= term ( V * V ) """
        tree.value = _convertstr(tree[0].value) * _convertstr(tree[2].value)

    def p_term_div(self, tree):
        """ V ::= term ( V / V ) """
        tree.value = _convertstr(tree[0].value) / tree[2].value

    def p_value_paren(self, tree):
        """ V ::= value ( LPAREN V RPAREN ) """
        tree.value = _convertstr(tree[1].value)
        tree.svalue = f'({str(tree[1].value):s})'

    def p_arglist(self, tree):
        """ V ::= arglist ( V , V ) """
        if isinstance(tree[0].value, list):
            tree.value = tree[0].value + [tree[2].value]
        else:
            tree.value = [tree[0].value, tree[2].value]
        try:
            tree.svalue = f'{tree[0].svalue:s},{tree[2].svalue:s}'
        except AttributeError:
            pass  # We only care about this for relatively simple constructs.

    @staticmethod
    def _get_names_from_tree_values(args):
        names = []
        for arg in args:
            if hasattr(arg, 'meta') and 'expr' in arg.meta:
                names.append(arg.meta['expr'])
            else:
                names.append(str(arg))
        return f"({','.join(names)})"

    def p_functioncall(self, tree):
        # Where all the real interpreter action is.
        # Note that things that should only be done at the top level
        # are performed in :func:`interpret` defined below.
        """ V ::= function_call ( V LPAREN V RPAREN ) """
        if not isinstance(tree[2].value, list):
            args = [tree[2].value]
        else:
            args = tree[2].value

        fname = tree[0].value
        metadata = {'expr': f'{fname}{self._get_names_from_tree_values(args)}'}

        if fname not in _SYFUNCTIONS:
            log.error(f'Unknown function: {fname}')
            self.error(fname)

        else:
            # Constant spectrum
            if fname == 'unit':
                if args[1] not in _SYFORMS:
                    log.error(f'Unrecognized unit: {args[1]}')
                    self.error(fname)
                try:
                    fluxunit = units.validate_unit(args[1])
                    tree.value = SourceSpectrum(
                        ConstFlux1D, amplitude=args[0]*fluxunit, meta=metadata)
                except NotImplementedError as e:
                    log.error(str(e))
                    self.error(fname)

            # Black body
            elif fname == 'bb':
                tree.value = SourceSpectrum(
                    BlackBodyNorm1D, temperature=args[0]*u.K)

            # Power law
            elif fname == 'pl':
                if args[2] not in _SYFORMS:
                    log.error(f'Unrecognized unit: {args[2]}')
                    self.error(fname)
                try:
                    fluxunit = units.validate_unit(args[2])
                    tree.value = SourceSpectrum(
                        PowerLawFlux1D, amplitude=1*fluxunit, x_0=args[0]*u.AA,
                        alpha=-args[1], meta=metadata)
                except (synexceptions.SynphotError, NotImplementedError) as e:
                    log.error(str(e))
                    self.error(fname)

            # Box throughput
            elif fname == 'box':
                tree.value = SpectralElement(
                    Box1D, amplitude=1, x_0=args[0]*u.AA, width=args[1]*u.AA,
                    meta=metadata)

            # Source spectrum from file
            elif fname == 'spec':
                tree.value = SourceSpectrum.from_file(irafconvert(args[0]))
                tree.value.meta.update(metadata)

            # Passband
            elif fname == 'band':
                tree.value = spectrum.band(tree[2].svalue)  # string value
                tree.value.meta.update(metadata)

            # Gaussian emission line
            elif fname == 'em':
                if args[3] not in _SYFORMS:
                    log.error(f'Unrecognized unit: {args[3]}')
                    self.error(fname)
                x0 = args[0] * u.AA
                fluxunit = units.validate_unit(args[3])
                totflux = args[2] * (fluxunit * u.AA)
                tree.value = SourceSpectrum(
                    GaussianFlux1D, total_flux=totflux, mean=x0,
                    fwhm=args[1]*u.AA)

            # Catalog interpolation
            elif fname == 'icat':
                tree.value = grid_to_spec(*args)

            # Renormalize source spectrum
            elif fname == 'rn':
                sp = args[0]
                bp = args[1]
                fluxunit = units.validate_unit(args[3])
                rnval = args[2] * fluxunit

                if not isinstance(sp, SourceSpectrum):
                    sp = SourceSpectrum.from_file(irafconvert(sp))

                if not isinstance(bp, SpectralElement):
                    bp = SpectralElement.from_file(irafconvert(bp))

                # Always force the renormalization to occur: prevent exceptions
                # in case of partial overlap. Less robust but duplicates
                # IRAF SYNPHOT. Force the renormalization in the case of
                # partial overlap, but raise an exception if the spectrum and
                # bandpass are entirely disjoint.
                try:
                    tree.value = sp.normalize(
                        rnval, band=bp, area=conf.area, vegaspec=spectrum.Vega)
                except synexceptions.PartialOverlap:
                    tree.value = sp.normalize(
                        rnval, band=bp, area=conf.area, vegaspec=spectrum.Vega,
                        force=True)
                    tree.value.warnings = {
                        'force_renorm': ('Renormalization exceeds the limit '
                                         'of the specified passband.')}
                tree.value.meta.update(metadata)

            # Redshift source spectrum (flat spectrum if fails)
            elif fname == 'z':
                sp = args[0]

                # ETC generates junk (i.e., 'null') sometimes
                if isinstance(sp, str) and sp != 'null':
                    sp = SourceSpectrum.from_file(irafconvert(sp))

                if isinstance(sp, SourceSpectrum):
                    tree.value = sp
                    tree.value.z = args[1]
                else:
                    tree.value = SourceSpectrum(
                        ConstFlux1D, amplitude=1*units.PHOTLAM)

                tree.value.meta.update(metadata)

            # Extinction
            elif fname == 'ebmvx':
                try:
                    tree.value = spectrum.ebmvx(args[1], args[0])
                except synexceptions.SynphotError as e:
                    log.error(str(e))
                    self.error(fname)
                tree.value.meta.update(metadata)

            # Default
            else:
                tree.value = (f'would call {fname} with the following args: '
                              f'{repr(args)}')


def tokens_info(tlist):  # pragma: no cover
    """Print tokens for debugging.

    Parameters
    ----------
    tlist : list
        List of tokens.

    """
    for token in tlist:
        log.info(f'{token.type} {token.attr}')


def scan(input_str):
    """Scan language string."""
    scanner = Scanner()
    input_str = input_str.replace('%2b', '+')
    return scanner.tokenize(input_str)


def parse(tokens):
    """Parse tokens."""
    parser = BaseParser(AST)
    return parser.parse(tokens)


def interpret(ast):
    """Interpret AST."""
    interpreter = Interpreter(ast)
    interpreter.match()
    value = ast.value
    return _convertstr(value)


def parse_spec(syncommand):
    """Parse a classic SYNPHOT command and return the resulting spectrum.

    Parameters
    ----------
    syncommand : str
        SYNPHOT command string.

    Returns
    -------
    sp : obj
        Spectrum object.

    """
    return interpret(parse(scan(syncommand)))
